#!/usr/bin/env python
# -*- coding: utf8 -*-

# ============================================================================
#  Copyright (c) nexB Inc. http://www.nexb.com/ - All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ============================================================================

import io
import os
import unittest

from testing_utils import get_test_loc
from testing_utils import get_temp_file

from attributecode import INFO
from attributecode import attrib
from attributecode import gen
from attributecode import model


class TemplateTest(unittest.TestCase):

    def test_check_template_simple_valid_returns_None(self):
        expected = None
        assert expected == attrib.check_template('template_string')

    def test_check_template_complex_valid_returns_None(self):
        template = '''
        {% for about in abouts -%}
            {{ about.name.value }}: {{ about.version.value }}
            {% for res in about.about_resource.value -%}
                resource: {{ res }}
            {% endfor -%}
        {% endfor -%}'''
        expected = None
        assert expected == attrib.check_template(template)

    def test_check_template_complex_invalid_returns_error(self):
        template = '''
        {% for about in abouts -%}
            {{ about.name.value }}: {{ about.version.value }}
            {% for res in about.about_ressdsdsdsdsdsdource.value -%}
                resource: {{] res }}
            {% endfor -%}
        {% endfor -%}'''
        expected = (5, "unexpected ']'")
        assert expected == attrib.check_template(template)

    def test_check_template_invalid_return_error_lineno_and_message(self):
        expected = 1, "unexpected end of template, expected 'end of print statement'."
        assert expected == attrib.check_template('{{template_string')

    def test_check_template_all_builtin_templates_are_valid(self):
        builtin_templates_dir = os.path.dirname(attrib.DEFAULT_TEMPLATE_FILE)
        for template in os.listdir(builtin_templates_dir):
            template_loc = os.path.join(builtin_templates_dir, template)
            with open(template_loc, 'r', encoding='utf-8', errors='replace') as tmpl:
                template = tmpl.read()
            try:
                assert None == attrib.check_template(template)
            except:
                raise Exception(template_loc)


class GenerateTest(unittest.TestCase):

    def test_generate_from_collected_inventory_wih_custom_temaplte(self):
        test_file = get_test_loc('test_attrib/gen_simple/attrib.ABOUT')
        errors, abouts = model.collect_inventory(test_file)
        assert not errors

        test_template = get_test_loc('test_attrib/gen_simple/test.template')
        with open(test_template) as tmpl:
            template = tmpl.read()

        expected = (
            'Apache HTTP Server: 2.4.3\n'
            'resource: httpd-2.4.3.tar.gz\n')

        license_dict = {}
        is_about_input = True
        min_license_score=0
        scancode = False
        error, result = attrib.generate(abouts, is_about_input, license_dict, scancode, min_license_score, template=template)
        assert expected == result
        assert not error

    def test_generate_with_default_template(self):
        test_file = get_test_loc('test_attrib/gen_default_template/attrib.ABOUT')
        errors, abouts = model.collect_inventory(test_file)
        assert not errors

        license_dict = {}
        is_about_input = True
        min_license_score=0
        scancode = False

        error, result = attrib.generate_from_file(abouts, is_about_input, license_dict, scancode, min_license_score)
        assert not error

        expected_file = get_test_loc(
            'test_attrib/gen_default_template/expected_default_attrib.html')
        with open(expected_file) as exp:
            expected = exp.read()

        # strip the timestamp: the timestamp is wrapped in italic block
        result = remove_timestamp(result)
        expected = remove_timestamp(expected)
        # Ignore all white spaces and newline
        result = result.replace('\n', '').replace(' ', '')
        expected = expected.replace('\n', '').replace(' ', '')
        assert expected == result

    def test_lic_key_name_sync(self):
        test_file = get_test_loc('test_attrib/gen_license_key_name_check/test.ABOUT')
        expected = get_test_loc('test_attrib/gen_license_key_name_check/expected/expected.html')
        template_loc = get_test_loc('test_attrib/gen_license_key_name_check/custom.template')
        output_file = get_temp_file()

        license_dict = {}
        is_about_input = True

        errors, abouts = model.collect_inventory(test_file)
        attrib.generate_and_save(abouts, is_about_input, license_dict, output_file, template_loc=template_loc)

        with open(output_file) as of:
            f1 = '\n'.join(of.readlines(False))
        with open(expected) as ef:
            f2 = '\n'.join(ef.readlines(False))

        assert f1 == f2

    def test_scancode_input(self):
        test_file = get_test_loc('test_attrib/scancode_input/clean-text-0.3.0-mod-lceupi.json')
        errors, abouts = gen.load_inventory(test_file, scancode=True)
        # No validation is done for the scancode input as it usually contains duplicated entry of
        # detected licenses which is not allow in the current spec.
        #expected_errors = [(40, 'Field about_resource: Unable to verify path: isc_lic.py: No base directory provided')]
        result = [(level, e) for level, e in errors if level > INFO]
        #assert expected_errors == result
        assert result == []

        lic_dict = {'isc': ['ISC License',
                            'isc.LICENSE',
                            'Permission to use, copy, modify, and/or distribute this software for any purpose\nwith or without fee is hereby granted, provided that the above copyright notice\nand this permission notice appear in all copies.\n\nTHE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH\nREGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND\nFITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,\nINDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS\nOF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER\nTORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF\nTHIS SOFTWARE.\n',
                            'https://scancode-licensedb.aboutcode.org/isc.LICENSE'],
                    'mit': ['MIT License',
                            'mit.LICENSE',
                            'Permission is hereby granted, free of charge, to any person obtaining\na copy of this software and associated documentation files (the\n"Software"), to deal in the Software without restriction, including\nwithout limitation the rights to use, copy, modify, merge, publish,\ndistribute, sublicense, and/or sell copies of the Software, and to\npermit persons to whom the Software is furnished to do so, subject to\nthe following conditions:\n\nThe above copyright notice and this permission notice shall be\nincluded in all copies or substantial portions of the Software.\n\nTHE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,\nEXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF\nMERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.\nIN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY\nCLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,\nTORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE\nSOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.',
                            'https://scancode-licensedb.aboutcode.org/mit.LICENSE']}
        is_about_input = False
        scancode = True
        errors, result = attrib.generate_from_file(abouts, is_about_input, lic_dict, scancode, min_license_score=0)
        expected_errors = []
        #result = [(level, e) for level, e in errors if level > INFO]
        #assert expected_errors == result
        assert not errors

        expected_file = get_test_loc(
            'test_attrib/scancode_input/expect.html')
        with open(expected_file) as exp:
            expected = exp.read()

        # strip the timestamp: the timestamp is wrapped in italic block
        result = remove_timestamp(result)
        expected = remove_timestamp(expected)
        # For whatever reasons, the directly comparison between the result and the
        # expected doesn't work well, it works after removed all the newline and spaces
        #assert expected == result
        #assert expected.splitlines(False) == result.splitlines(False)
        assert expected.replace('\n','').replace(' ','').replace('\t','') == result.replace('\n','').replace(' ','').replace('\t','')

    def test_generate_with_csv(self):
        test_file = get_test_loc('test_attrib/default_template/simple_sample.csv')
        errors, abouts = gen.load_inventory(test_file)

        lic_dict = {'isc': ['ISC License',
                            'isc.LICENSE',
                            'Permission to use, copy, modify, and/or distribute this software for any purpose\nwith or without fee is hereby granted, provided that the above copyright notice\nand this permission notice appear in all copies.\n\nTHE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH\nREGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND\nFITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,\nINDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS\nOF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER\nTORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF\nTHIS SOFTWARE.\n',
                            'https://scancode-licensedb.aboutcode.org/isc.LICENSE']}
        is_about_input = False
        scancode = False

        error, result = attrib.generate_from_file(abouts, is_about_input, lic_dict, scancode, min_license_score=0)
        assert not error

        expected_file = get_test_loc(
            'test_attrib/default_template/expect.html')
        with open(expected_file) as exp:
            expected = exp.read()

        # strip the timestamp: the timestamp is wrapped in italic block
        result = remove_timestamp(result)
        expected = remove_timestamp(expected)
        #assert expected == result
        assert expected.replace('\n','').replace(' ','') == result.replace('\n','').replace(' ','')

def remove_timestamp(html_text):
    """
    Return the `html_text` generated attribution stripped from timestamps: the
    timestamp is wrapped in italic block in the default template.
    """
    return '\n'.join(x for x in html_text.splitlines() if not '<i>' in x)



#!/usr/bin/env python
# -*- coding: utf8 -*-
# ============================================================================
#  Copyright (c) nexB Inc. http://www.nexb.com/ - All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ============================================================================

from collections import OrderedDict

import codecs
import csv
import json
import ntpath
import openpyxl
import os
import posixpath
import re
import shutil
import string
import sys
from distutils.dir_util import copy_tree
from itertools import zip_longest

from attributecode import CRITICAL
from attributecode import WARNING
from attributecode import Error

on_windows = 'win32' in sys.platform

# boolean field name
boolean_fields = ['redistribute', 'attribute', 'track_change', 'modified', 'internal_use_only']
file_fields = ['about_resource', 'notice_file', 'changelog_file', 'author_file']


def to_posix(path):
    """
    Return a path using the posix path separator given a path that may contain
    posix or windows separators, converting "\\" to "/". NB: this path will
    still be valid in the windows explorer (except for a UNC or share name). It
    will be a valid path everywhere in Python. It will not be valid for windows
    command line operations.
    """
    return path.replace(ntpath.sep, posixpath.sep)


UNC_PREFIX = u'\\\\?\\'
UNC_PREFIX_POSIX = to_posix(UNC_PREFIX)
UNC_PREFIXES = (UNC_PREFIX_POSIX, UNC_PREFIX,)

valid_file_chars = string.digits + string.ascii_letters + '_-.+()~[]{}|@%' + ' '


def invalid_chars(path):
    """
    Return a list of invalid characters in the file name of `path`.
    """
    path = to_posix(path)
    rname = resource_name(path)
    name = rname.lower()
    return [c for c in name if c not in valid_file_chars]


def check_file_names(paths):
    """
    Given a sequence of file paths, check that file names are valid and that
    there are no case-insensitive duplicates in any given directories.
    Return a list of errors.

    From spec :
        A file name can contain only these US-ASCII characters:
        - digits from 0 to 9
        - uppercase and lowercase letters from A to Z
        - the _ underscore, - dash and . period signs.
    From spec:
     The case of a file name is not significant. On case-sensitive file
     systems (such as Linux), a tool must raise an error if two ABOUT files
     stored in the same directory have the same lowercase file name.
    """
    # FIXME: this should be a defaultdicts that accumulates all duplicated paths
    seen = {}
    errors = []
    for orig_path in paths:
        path = orig_path
        invalid = invalid_chars(path)
        if invalid:
            invalid = ''.join(invalid)
            msg = ('Invalid characters %(invalid)r in file name at: '
                   '%(path)r' % locals())
            errors.append(Error(CRITICAL, msg))

        path = to_posix(orig_path)
        name = resource_name(path).lower()
        parent = posixpath.dirname(path)
        path = posixpath.join(parent, name)
        path = posixpath.normpath(path)
        path = posixpath.abspath(path)
        existing = seen.get(path)
        if existing:
            msg = ('Duplicate files: %(orig_path)r and %(existing)r '
                   'have the same case-insensitive file name' % locals())
            errors.append(Error(CRITICAL, msg))
        else:
            seen[path] = orig_path
    return errors


def wrap_boolean_value(context):
    updated_context = ''
    for line in context.splitlines():
        """
        wrap the boolean value in quote
        """
        key = line.partition(':')[0]
        value = line.partition(':')[2].strip()
        value = '"' + value + '"'
        if key in boolean_fields and not value == "":
            updated_context += key + ': ' + value + '\n'
        else:
            updated_context += line + '\n'
    return updated_context


def replace_tab_with_spaces(context):
    updated_context = ''
    for line in context.splitlines():
        """
        Replace tab with 4 spaces
        """
        updated_context += line.replace('\t', '    ') + '\n'
    return updated_context


# TODO: rename to normalize_path
def get_absolute(location):
    """
    Return an absolute normalized location.
    """
    location = os.path.expanduser(location)
    location = os.path.expandvars(location)
    location = os.path.normpath(location)
    location = os.path.abspath(location)
    return location


def get_locations(location):
    """
    Return a list of locations of files given the `location` of a
    a file or a directory tree containing ABOUT files.
    File locations are normalized using posix path separators.
    """
    location = add_unc(location)
    location = get_absolute(location)
    assert os.path.exists(location)

    if os.path.isfile(location):
        yield location
    else:
        for base_dir, _, files in os.walk(location):
            for name in files:
                bd = to_posix(base_dir)
                yield posixpath.join(bd, name)


def get_about_locations(location):
    """
    Return a list of locations of ABOUT files given the `location` of a
    a file or a directory tree containing ABOUT files.
    File locations are normalized using posix path separators.
    """
    for loc in get_locations(location):
        if is_about_file(loc):
            yield loc


def norm(p):
    """
    Normalize the path
    """
    if p.startswith(UNC_PREFIX) or p.startswith(to_posix(UNC_PREFIX)):
        p = p.strip(UNC_PREFIX).strip(to_posix(UNC_PREFIX))
    p = to_posix(p)
    p = p.strip(posixpath.sep)
    p = posixpath.normpath(p)
    return p


def get_relative_path(base_loc, full_loc):
    """
    Return a posix path for a given full location relative to a base location.
    The first segment of the different between full_loc and base_loc will become
    the first segment of the returned path.
    """
    base = norm(base_loc)
    path = norm(full_loc)

    assert path.startswith(base), ('Cannot compute relative path: '
                                   '%(path)r does not start with %(base)r'
                                   % locals())
    base_name = resource_name(base)
    no_dir = base == base_name
    same_loc = base == path
    if same_loc:
        # this is the case of a single file or single dir
        if no_dir:
            # we have no dir: the full path is the same as the resource name
            relative = base_name
        else:
            # we have at least one dir
            parent_dir = posixpath.dirname(base)
            parent_dir = resource_name(parent_dir)
            relative = posixpath.join(parent_dir, base_name)
    else:
        relative = path[len(base) + 1:]
        # We don't want to keep the first segment of the root of the returned path.
        # See https://github.com/nexB/attributecode/issues/276
        # relative = posixpath.join(base_name, relative)
    return relative


def to_native(path):
    """
    Return a path using the current OS path separator given a path that may
    contain posix or windows separators, converting "/" to "\\" on windows
    and "\\" to "/" on posix OSes.
    """
    path = path.replace(ntpath.sep, os.path.sep)
    path = path.replace(posixpath.sep, os.path.sep)
    return path


def is_about_file(path):
    """
    Return True if the path represents a valid ABOUT file name.
    """
    if path:
        path = path.lower()
        return path.endswith('.about') and path != '.about'


def resource_name(path):
    """
    Return the file or directory name from a path.
    """
    path = path.strip()
    path = to_posix(path)
    path = path.rstrip(posixpath.sep)
    _left, right = posixpath.split(path)
    return right.strip()


def load_csv(location):
    """
    Read CSV at `location`, return a list of ordered dictionaries, one
    for each row.
    """
    results = []
    with open(location, mode='r', encoding='utf-8-sig',
                     errors='replace') as csvfile:
        for row in csv.DictReader(csvfile):
            # convert all the column keys to lower case
            updated_row = {key.lower().strip(): value for key, value in row.items()}
            results.append(updated_row)
    return results

def load_json(location):
    """
    Read JSON file at `location` and return a list of ordered dicts, one for
    each entry.
    """
    with open(location) as json_file:
        results = json.load(json_file)

    # FIXME: this is too clever and complex... IMHO we should not try to guess the format.
    # instead a command line option should be provided explictly to say what is the format
    if isinstance(results, list):
        results = sorted(results)
    else:
        if u'aboutcode_manager_notice' in results:
            results = results['components']
        elif u'scancode_notice' in results:
            results = results['files']
        else:
            results = [results]
    return results


# FIXME: rename to is_online: BUT do we really need this at all????
def have_network_connection():
    """
    Return True if an HTTP connection to some public web site is possible.
    """
    import socket
    import http.client as httplib

    http_connection = httplib.HTTPConnection('dejacode.org', timeout=10)  # NOQA
    try:
        http_connection.connect()
    except socket.error:
        return False
    else:
        return True


def extract_zip(location):
    """
    Extract a zip file at location in a temp directory and return the temporary
    directory where the archive was extracted.
    """
    import zipfile
    import tempfile

    if not zipfile.is_zipfile(location):
        raise Exception('Incorrect zip file %(location)r' % locals())

    archive_base_name = os.path.basename(location).replace('.zip', '')
    base_dir = tempfile.mkdtemp(prefix='aboutcode-toolkit-extract-')
    target_dir = os.path.join(base_dir, archive_base_name)
    target_dir = add_unc(target_dir)
    os.makedirs(target_dir)

    if target_dir.endswith((ntpath.sep, posixpath.sep)):
        target_dir = target_dir[:-1]

    with zipfile.ZipFile(location) as zipf:
        for info in zipf.infolist():
            name = info.filename
            content = zipf.read(name)
            target = os.path.join(target_dir, name)
            is_dir = target.endswith((ntpath.sep, posixpath.sep))
            if is_dir:
                target = target[:-1]
            parent = os.path.dirname(target)
            if on_windows:
                target = target.replace(posixpath.sep, ntpath.sep)
                parent = parent.replace(posixpath.sep, ntpath.sep)
            if not os.path.exists(parent):
                os.makedirs(add_unc(parent))
            if not content and is_dir:
                if not os.path.exists(target):
                    os.makedirs(add_unc(target))
            if not os.path.exists(target):
                with open(target, 'wb') as f:
                    f.write(content)
    return target_dir


def add_unc(location):
    """
    Convert a `location` to an absolute Window UNC path to support long paths on
    Windows. Return the location unchanged if not on Windows. See
    https://msdn.microsoft.com/en-us/library/aa365247.aspx
    """
    if on_windows and not location.startswith(UNC_PREFIX):
        if location.startswith(UNC_PREFIX_POSIX):
            return UNC_PREFIX + os.path.abspath(location.strip(UNC_PREFIX_POSIX))
        return UNC_PREFIX + os.path.abspath(location)
    return location


# FIXME: add docstring
def copy_license_notice_files(fields, base_dir, reference_dir, afp):
    """
    Given a list of (key, value) `fields` tuples and a `base_dir` where ABOUT
    files and their companion LICENSe are store, and an extra `reference_dir`
    where reference license an notice files are stored and the `afp`
    about_file_path value, this function will copy to the base_dir the
    license_file or notice_file if found in the reference_dir
    """
    errors = []
    copy_file_name = ''
    for key, value in fields:
        if key == 'license_file' or key == 'notice_file':
            if value:
                # This is to handle multiple license_file value in CSV format
                # The following code will construct a list to contain the
                # license file(s) that need to be copied.
                # Note that *ONLY* license_file field allows \n. Others file
                # fields that have \n will prompts error at validation stage
                file_list = []
                if '\n' in value:
                    f_list = value.split('\n')
                else:
                    if not isinstance(value, list):
                        f_list = [value]
                    else:
                        f_list = value
                # The following code is to adopt the approach from #404
                # to use comma for multiple files which refer the same license
                for item in f_list:
                    if ',' in item:
                        item_list = item.split(',')
                        for i in item_list:
                            file_list.append(i.strip())
                    else:
                        file_list.append(item)
            else:
                continue

            for copy_file_name in file_list:
                from_lic_path = posixpath.join(to_posix(reference_dir), copy_file_name)
                about_file_dir = os.path.dirname(to_posix(afp)).lstrip('/')
                to_lic_path = posixpath.join(to_posix(base_dir), about_file_dir)
                if not os.path.exists(posixpath.join(to_lic_path, copy_file_name)):
                    err = copy_file(from_lic_path, to_lic_path)
                    if err:
                        errors.append(err)
    return errors


def copy_file(from_path, to_path):
    error = ''
    # Return if the from_path is empty or None.
    if not from_path:
        return

    if on_windows:
        if not from_path.startswith(UNC_PREFIXES):
            from_path = add_unc(from_path)
        if not to_path.startswith(UNC_PREFIXES):
            to_path = add_unc(to_path)

    # Strip the white spaces
    from_path = from_path.strip()
    to_path = to_path.strip()
    # Errors will be captured when doing the validation
    if not os.path.exists(from_path):
        return ''

    if not posixpath.exists(to_path):
        os.makedirs(to_path)
    try:
        if os.path.isdir(from_path):
            # Copy the whole directory structure
            if from_path.endswith('/'):
                from_path = from_path.rpartition('/')[0]
            folder_name = os.path.basename(from_path)
            to_path = os.path.join(to_path, folder_name)
            if os.path.exists(to_path):
                msg = to_path + ' is already existed and is replaced by ' + from_path
                error = Error(WARNING, msg)
            copy_tree(from_path, to_path)
        else:
            file_name = os.path.basename(from_path)
            to_file_path = os.path.join(to_path, file_name)
            if os.path.exists(to_file_path):
                msg = to_file_path + ' is already existed and is replaced by ' + from_path
                error = Error(WARNING, msg)
            shutil.copy2(from_path, to_path)
        return error
    except Exception as e:
        msg = 'Cannot copy file at %(from_path)r.' % locals()
        error = Error(CRITICAL, msg)
        return error


# FIXME: we should use a license object instead
def ungroup_licenses(licenses):
    """
    Ungroup multiple licenses information
    """
    lic_key = []
    lic_name = []
    lic_file = []
    lic_url = []
    spdx_lic_key = []
    lic_score = []
    lic_matched_text = []
    for lic in licenses:
        if 'key' in lic:
            lic_key.append(lic['key'])
        if 'name' in lic:
            lic_name.append(lic['name'])
        if 'file' in lic:
            lic_file.append(lic['file'])
        if 'url' in lic:
            lic_url.append(lic['url'])
        if 'spdx_license_key' in lic:
            spdx_lic_key.append(lic['spdx_license_key'])
        if 'score' in lic:
            lic_score.append(lic['score'])
        if 'matched_text' in lic:
            lic_matched_text.append(lic['matched_text'])
    return lic_key, lic_name, lic_file, lic_url, spdx_lic_key, lic_score, lic_matched_text


# FIXME: add docstring
def format_about_dict_output(about_dictionary_list):
    formatted_list = []
    for element in about_dictionary_list:
        row_list = dict()
        for key in element:
            if element[key]:
                if isinstance(element[key], list):
                    row_list[key] = u'\n'.join((element[key]))
                elif key == u'about_resource':
                    row_list[key] = u'\n'.join((element[key].keys()))
                else:
                    row_list[key] = element[key]
        formatted_list.append(row_list)
    return formatted_list


# FIXME: add docstring
def format_about_dict_for_json_output(about_dictionary_list):
    licenses = ['license_key', 'license_name', 'license_file', 'license_url']
    json_formatted_list = []
    for element in about_dictionary_list:
        row_list = dict()
        # FIXME: aboid using parallel list... use an object instead
        license_key = []
        license_name = []
        license_file = []
        license_url = []

        for key in element:
            if element[key]:
                # The 'about_resource' is an ordered dict
                if key == 'about_resource':
                    row_list[key] = list(element[key].keys())[0]
                elif key in licenses:
                    if key == 'license_key':
                        license_key = element[key]
                    elif key == 'license_name':
                        license_name = element[key]
                    elif key == 'license_file':
                        license_file = element[key]
                    elif key == 'license_url':
                        license_url = element[key]
                else:
                    row_list[key] = element[key]

        # Group the same license information in a list
        license_group = list(zip_longest(license_key, license_name, license_file, license_url))
        if license_group:
            licenses_list = []
            for lic_group in license_group:
                lic_dict = dict()
                if lic_group[0]:
                    lic_dict['key'] = lic_group[0]
                if lic_group[1]:
                    lic_dict['name'] = lic_group[1]
                if lic_group[2]:
                    lic_dict['file'] = lic_group[2]
                if lic_group[3]:
                    lic_dict['url'] = lic_group[3]
                licenses_list.append(lic_dict)
            row_list['licenses'] = licenses_list
        json_formatted_list.append(row_list)
    return json_formatted_list


def unique(sequence):
    """
    Return a list of unique items found in sequence. Preserve the original
    sequence order.
    For example:
    >>> unique([1, 5, 3, 5])
    [1, 5, 3]
    """
    deduped = []
    for item in sequence:
        if item not in deduped:
            deduped.append(item)
    return deduped


def filter_errors(errors, minimum_severity=WARNING):
    """
    Return a list of unique `errors` Error object filtering errors that have a
    severity below `minimum_severity`.
    """
    return [e for e in errors if e.severity >= minimum_severity]


def create_dir(location):
    """
    Create directory or directory tree at location, ensuring it is readable
    and writeable.
    """
    import stat
    if not os.path.exists(location):
        os.makedirs(location)
        os.chmod(location, stat.S_IRWXU | stat.S_IRWXG
                 | stat.S_IROTH | stat.S_IXOTH)


def get_temp_dir(sub_dir_path=None):
    """
    Create a unique new temporary directory location. Create directories
    identified by sub_dir_path if provided in this temporary directory.
    Return the location for this unique directory joined with the
    sub_dir_path if any.
    """
    new_temp_dir = build_temp_dir()

    if sub_dir_path:
        # create a sub directory hierarchy if requested
        new_temp_dir = os.path.join(new_temp_dir, sub_dir_path)
        create_dir(new_temp_dir)
    return new_temp_dir


def build_temp_dir(prefix='attributecode-'):
    """
    Create and return a new unique empty directory created in base_dir.
    """
    import tempfile
    location = tempfile.mkdtemp(prefix=prefix)
    create_dir(location)
    return location

def get_file_text(file_name, reference):
    """
    Return the file content from the license_file/notice_file field from the
    given reference directory.
    """
    error = ''
    text = ''
    file_path = os.path.join(reference, file_name)
    if not os.path.exists(file_path):
        msg = "The file " + file_path + " does not exist"
        error = Error(CRITICAL, msg)
    else:
        with codecs.open(file_path, 'rb', encoding='utf-8-sig', errors='replace') as txt:
        #with io.open(file_path, encoding='utf-8') as txt:
            text = txt.read()
    return error, text

def convert_object_to_dict(about):
    """
    Convert the list of field object
        [Field(name='name', value=''), Field(name='version', value='')]
    to a dictionary
    """
    about_dict = {}
    # Convert all the supported fields into a dictionary
    fields_dict = getattr(about, 'fields')
    custom_fields_dict = getattr(about, 'custom_fields')
    supported_dict = {**fields_dict, **custom_fields_dict}
    for field in supported_dict:
        key = supported_dict[field].name
        value = supported_dict[field].value
        about_dict[key] = value
    return about_dict

def load_scancode_json(location):
    """
    Read the scancode JSON file at `location` and return a list of dictionaries.
    """
    updated_results = []

    with open(location) as json_file:
        results = json.load(json_file)
    results = results['files']
    # Rename the "path" to "about_resource"
    for item in results:
        updated_dict = {}
        for key in item:
            if key == 'path':
                updated_dict['about_resource'] = item[key]
            else:
                updated_dict[key] = item[key]
        updated_results.append(updated_dict)
    return updated_results

def load_excel(location):
    """
    Read XLSX at `location`, return a list of ordered dictionaries, one
    for each row.
    """
    results = []
    errors = []
    import warnings

    # This is to prevent showing the: warn("Workbook contains no default style, apply openpyxl's default")
    with warnings.catch_warnings(record=True):
        sheet_obj = openpyxl.load_workbook(location).active
    max_col = sheet_obj.max_column

    index = 1
    col_keys = []
    mapping_dict = {}

    while index <= max_col:
        value = sheet_obj.cell(row=1, column=index).value
        if value in col_keys:
            msg = 'Duplicated column name, ' + str(value) + ', detected.'
            errors.append(Error(CRITICAL, msg))
            return errors, results
        if value in mapping_dict:
            value = mapping_dict[value]
        col_keys.append(value)
        index = index + 1

    for row in sheet_obj.iter_rows(min_row=2, values_only=True):
        row_dict = OrderedDict()
        index = 0
        while index < max_col:
            value = row[index]
            if value:
                row_dict[col_keys[index]] = value
            else:
                row_dict[col_keys[index]] = ''
            index = index + 1
        results.append(row_dict)
    return errors, results

def write_licenses(lic_dict, location):
    import io

    loc = to_posix(location)
    errors = []

    if not posixpath.exists(loc):
        os.makedirs(add_unc(loc))
    try:
        for lic in lic_dict:
            output_location = posixpath.join(loc, lic)
            with open(output_location, 'w', encoding='utf-8', errors='replace') as out:
                out.write(lic_dict[lic])
    except Exception as e:
        msg = str(e)
        errors.append(Error(CRITICAL, msg))
    return errors


"""
Return True if a string s  name is safe to use as an attribute name.
"""
is_valid_name = re.compile(r'^[A-Za-z_][A-Za-z0-9_]*$').match



#!/usr/bin/env python
# -*- coding: utf8 -*-

# ============================================================================
#  Copyright (c) nexB Inc. http://www.nexb.com/ - All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ============================================================================

import io
import json
import os
import posixpath
import shutil
import unittest
from unittest import mock

import saneyaml

from attributecode import CRITICAL
from attributecode import ERROR
from attributecode import INFO
from attributecode import WARNING
from attributecode import Error
from attributecode import model
from attributecode.util import add_unc, norm, on_windows
from attributecode.util import load_csv
from attributecode.util import to_posix
from attributecode.util import replace_tab_with_spaces

from testing_utils import extract_test_loc
from testing_utils import get_temp_dir
from testing_utils import get_temp_file
from testing_utils import get_test_loc


def check_csv(expected, result, regen=False, fix_cell_linesep=False):
    """
    Assert that the contents of two CSV files locations `expected` and
    `result` are equal.
    """
    if regen:
        shutil.copyfile(result, expected)
    expected = sorted([sorted(d.items()) for d in load_csv(expected)])
    result = [d.items() for d in load_csv(result)]
    if fix_cell_linesep:
        result = [list(fix_crlf(items)) for items in result]
    result = sorted(sorted(items) for items in result)

    assert expected == result


def fix_crlf(items):
    """
    Hackish... somehow the CVS returned on Windows is sometimes using a backward
    linesep convention:
    instead of LF inside cells and CRLF at EOL,
    they use CRLF everywhere.
    This is fixing this until we find can why
    """
    for key, value in items:
        if isinstance(value, str) and '\r\n' in value:
            value = value.replace('\r\n', '\n')
        yield key, value


def check_json(expected, result):
    """
    Assert that the contents of two JSON files are equal.
    """
    with open(expected) as e:
        expected = json.load(e, object_pairs_hook=dict)
    with open(result) as r:
        result = json.load(r, object_pairs_hook=dict)
    assert expected == result


def get_test_content(test_location):
    """
    Read file at test_location and return a unicode string.
    """
    return get_unicode_content(get_test_loc(test_location))


def get_unicode_content(location):
    """
    Read file at location and return a unicode string.
    """
    with open(location, encoding='utf-8', errors='replace') as doc:
        return doc.read()


class FieldTest(unittest.TestCase):

    def test_Field_init(self):
        model.Field()
        model.StringField()
        model.ListField()
        model.UrlField()
        model.BooleanField()
        model.PathField()
        model.FileTextField()
        model.PackageUrlField()

    def test_empty_Field_has_no_content(self):
        field = model.Field()
        assert not field.has_content

    def test_empty_Field_has_default_value(self):
        field = model.Field()
        assert '' == field.value

    def test_PathField_check_location(self):
        test_file = 'license.LICENSE'
        field = model.PathField(name='f', value=test_file, present=True)
        base_dir = get_test_loc('test_model/base_dir')

        errors = field.validate(base_dir=base_dir)
        expected_errrors = []
        assert expected_errrors == errors

        result = field.value[test_file]
        expected = add_unc(posixpath.join(to_posix(base_dir), test_file))
        assert expected == result

    def test_PathField_check_missing_location(self):
        test_file = 'does.not.exist'
        field = model.PathField(name='f', value=test_file, present=True)
        base_dir = get_test_loc('test_model/base_dir')
        errors = field.validate(base_dir=base_dir)

        file_path = posixpath.join(base_dir, test_file)
        err_msg = 'Field f: Path %s not found' % file_path

        expected_errors = [
            Error(CRITICAL, err_msg)]
        assert expected_errors == errors

        result = field.value[test_file]
        assert None == result

    def test_TextField_loads_file(self):
        field = model.FileTextField(
            name='f', value='license.LICENSE', present=True)

        base_dir = get_test_loc('test_model/base_dir')
        errors = field.validate(base_dir=base_dir)
        assert [] == errors

        expected = {'license.LICENSE': 'some license text'}
        assert expected == field.value

    def test_PackageUrlField_is_valid_url(self):
        assert model.PackageUrlField.is_valid_purl('pkg:pypi/saneyaml@0.1')

    def test_PackageUrlField_is_valid_url_no_version(self):
        assert model.PackageUrlField.is_valid_purl('pkg:pypi/saneyaml')

    def test_UrlField_is_valid_url(self):
        assert model.UrlField.is_valid_url('http://www.google.com')

    def test_UrlField_is_valid_url_not_starting_with_www(self):
        assert model.UrlField.is_valid_url('https://nexb.com')
        assert model.UrlField.is_valid_url('http://archive.apache.org/dist/httpcomponents/commons-httpclient/2.0/source/commons-httpclient-2.0-alpha2-src.tar.gz')
        assert model.UrlField.is_valid_url('http://de.wikipedia.org/wiki/Elf (Begriffsklärung)')
        assert model.UrlField.is_valid_url('http://nothing_here.com')

    def test_UrlField_is_valid_url_no_schemes(self):
        assert not model.UrlField.is_valid_url('google.com')
        assert not model.UrlField.is_valid_url('www.google.com')
        assert not model.UrlField.is_valid_url('')

    def test_UrlField_is_valid_url_not_ends_with_com(self):
        assert model.UrlField.is_valid_url('http://www.google')

    def test_UrlField_is_valid_url_ends_with_slash(self):
        assert model.UrlField.is_valid_url('http://www.google.co.uk/')

    def test_UrlField_is_valid_url_empty_URL(self):
        assert not model.UrlField.is_valid_url('http:')

    def check_validate(self, field_class, value, expected, expected_errors):
        """
        Check field values after validation
        """
        field = field_class(name='s', value=value, present=True)
        # check that validate can be applied multiple times without side effects
        for _ in range(2):
            errors = field.validate()
            assert expected_errors == errors
            assert expected == field.value

    def test_StringField_validate_trailing_spaces_are_removed(self):
        field_class = model.StringField
        value = 'trailin spaces  '
        expected = 'trailin spaces'
        self.check_validate(field_class, value, expected, expected_errors=[])

    def test_ListField_contains_list_after_validate(self):
        value = 'string'
        field_class = model.ListField
        expected = [value]
        self.check_validate(field_class, value, expected, expected_errors=[])

    def test_ListField_contains_stripped_strings_after_validate(self):
        value = '''first line
                   second line  '''
        field_class = model.ListField
        expected = ['first line', 'second line']
        self.check_validate(field_class, value, expected, expected_errors=[])

    def test_PathField_contains_stripped_strings_after_validate(self):
        value = '''first line
                   second line  '''
        field_class = model.ListField
        expected = ['first line', 'second line']
        self.check_validate(field_class, value, expected, expected_errors=[])

    def test_PathField_contains_dict_after_validate(self):
        value = 'string'
        field_class = model.PathField
        expected = dict([('string', None)])
        expected_errors = [
            Error(ERROR, 'Field s: Unable to verify path: string: No base directory provided')
                          ]
        self.check_validate(field_class, value, expected, expected_errors)

    def test_SingleLineField_has_errors_if_multiline(self):
        value = '''line1
        line2'''
        field_class = model.SingleLineField
        expected = value
        expected_errors = [Error(ERROR, 'Field s: Cannot span multiple lines: line1\n        line2')]
        self.check_validate(field_class, value, expected, expected_errors)


class YamlParseTest(unittest.TestCase):
    maxDiff = None

    def test_saneyaml_load_can_parse_simple_fields(self):
        test = get_test_content('test_model/parse/basic.about')
        result = saneyaml.load(test)

        expected = [
            ('single_line', 'optional'),
            ('other_field', 'value'),
        ]

        assert expected == list(result.items())

    def test_saneyaml_load_does_not_convert_to_crlf(self):
        test = get_test_content('test_model/crlf/about.ABOUT')
        result = saneyaml.load(test)

        expected = [
            (u'about_resource', u'.'),
            (u'name', u'pytest'),
            (u'description', u'first line\nsecond line\nthird line\n'),
            (u'copyright', u'copyright')
        ]
        assert expected == list(result.items())

    def test_saneyaml_load_can_parse_continuations(self):
        test = get_test_content('test_model/parse/continuation.about')
        result = saneyaml.load(test)

        expected = [
            ('single_line', 'optional'),
            ('other_field', 'value'),
            (u'multi_line', u'some value and more and yet more')
        ]

        assert expected == list(result.items())

    def test_saneyaml_load_can_handle_multiline_texts_and_strips_text_fields(self):
        test = get_test_content('test_model/parse/complex.about')
        result = saneyaml.load(test)

        expected = [
            ('single_line', 'optional'),
            ('other_field', 'value'),
            ('multi_line', 'some value and more and yet more'),
            ('yetanother', 'sdasd')]

        assert expected == list(result.items())

    def test_saneyaml_load_can_parse_verbatim_text_unstripped(self):
        test = get_test_content('test_model/parse/continuation_verbatim.about')
        result = saneyaml.load(test)

        expected = [
            (u'single_line', u'optional'),
            (u'other_field', u'value'),
            (u'multi_line', u'some value  \n  and more  \n      and yet more   \n \n')
        ]

        assert expected == list(result.items())

    def test_saneyaml_load_can_parse_verbatim_tab_text_unstripped(self):
        test = get_test_content('test_model/parse/continuation_verbatim_with_tab.about')
        data = replace_tab_with_spaces(test)
        result = saneyaml.load(data)

        expected = [
            (u'single_line', u'optional'),
            (u'other_field', u'value'),
            (u'multi_line', u'This is a long description\nwith tab.\n')
        ]

        assert expected == list(result.items())

    def test_saneyaml_load_report_error_for_invalid_field_name(self):
        test = get_test_content('test_model/parse/invalid_names.about')
        try:
            saneyaml.load(test)
            self.fail('Exception not raised')
        except Exception:
            pass

    def test_saneyaml_dangling_text_is_not_an_invalid_continuation(self):
        test = get_test_content('test_model/parse/invalid_continuation.about')
        result = saneyaml.load(test)
        expected = [
            (u'single_line', u'optional'),
            (u'other_field', u'value'),
            (u'multi_line', u'some value and more\ninvalid continuation2')
        ]
        assert expected == list(result.items())

    def test_saneyaml_load_accepts_unicode_keys_and_values(self):
        test = get_test_content('test_model/parse/non_ascii_field_name_value.about')
        result = saneyaml.load(test)
        expected = [
            ('name', 'name'),
            ('about_resource', '.'),
            ('owner', 'Matías Aguirre'),
            (u'Matías', u'unicode field name')
        ]
        assert expected == list(result.items())

    def test_saneyaml_load_accepts_blank_lines_and_spaces_in_field_names(self):
        test = '''
name: test space
version: 0.7.0
about_resource: about.py
field with spaces: This is a test case for field with spaces
'''

        result = saneyaml.load(test)

        expected = [
            ('name', 'test space'),
            ('version', '0.7.0'),
            ('about_resource', 'about.py'),
            (u'field with spaces', u'This is a test case for field with spaces'),
        ]

        assert expected == list(result.items())

    def test_saneyaml_loads_blank_lines_and_lines_without_no_colon(self):
        test = '''
name: no colon test
test
version: 0.7.0
about_resource: about.py
test with no colon
'''
        try:
            saneyaml.load(test)
            self.fail('Exception not raised')
        except Exception:
            pass


class AboutTest(unittest.TestCase):

    def test_About_load_ignores_original_field_order_and_uses_standard_predefined_order(self):
        # fields in this file are not in the standard order
        test_file = get_test_loc('test_model/parse/ordered_fields.ABOUT')
        a = model.About(test_file)
        assert [] == a.errors

        expected = ['about_resource', 'name', 'version', 'download_url']
        result = [f.name for f in a.all_fields() if f.present]
        assert expected == result

    def test_About_duplicate_field_names_are_detected_with_different_case(self):
        # This test is failing because the YAML does not keep the order when
        # loads the test files. For instance, it treat the 'About_Resource' as the
        # first element and therefore the dup key is 'about_resource'.
        test_file = get_test_loc('test_model/parse/dupe_field_name.ABOUT')
        a = model.About(test_file)
        expected = [
            Error(WARNING, 'Field About_Resource is a duplicate. Original value: "." replaced with: "new value"'),
            Error(WARNING, 'Field Name is a duplicate. Original value: "old" replaced with: "new"')
        ]

        result = a.errors
        assert sorted(expected) == sorted(result)

    def test_About_duplicate_field_names_are_not_reported_if_same_value(self):
        # This test is failing because the YAML does not keep the order when
        # loads the test files. For instance, it treat the 'About_Resource' as the
        # first element and therefore the dup key is 'about_resource'.
        test_file = get_test_loc('test_model/parse/dupe_field_name_no_new_value.ABOUT')
        a = model.About(test_file)
        expected = [
]
        result = a.errors
        assert sorted(expected) == sorted(result)

    def check_About_hydrate(self, about, fields):
        expected = set([
            'name',
            'homepage_url',
            'download_url',
            'version',
            'copyright',
            'date',
            'license_spdx',
            'license_text_file',
            'notice_file',
            'about_resource'])

        expected_errors = [
            Error(INFO, 'Custom Field: date'),
            Error(INFO, 'Custom Field: license_spdx'),
            Error(INFO, 'Custom Field: license_text_file')]

        errors = about.hydrate(fields)

        assert expected_errors == errors

        result = set([f.name for f in about.all_fields() if f.present])
        assert expected == result

    def test_About_hydrate_normalize_field_names_to_lowercase(self):
        test_content = get_test_content('test_gen/parser_tests/upper_field_names.ABOUT')
        fields = saneyaml.load(test_content).items()
        a = model.About()
        for _ in range(3):
            self.check_About_hydrate(a, fields)

    def test_About_with_existing_about_resource_has_no_error(self):
        test_file = get_test_loc('test_gen/parser_tests/about_resource_field.ABOUT')
        a = model.About(test_file)
        assert [] == a.errors
        result = a.about_resource.value['about_resource.c']
        # this means we have a location
        self.assertNotEqual([], result)

    def test_About_has_errors_when_about_resource_is_missing(self):
        test_file = get_test_loc('test_gen/parser_tests/.ABOUT')
        a = model.About(test_file)
        expected = [Error(CRITICAL, 'Field about_resource is required')]
        result = a.errors
        assert expected == result

    def test_About_has_errors_when_about_resource_does_not_exist(self):
        test_file = get_test_loc('test_gen/parser_tests/missing_about_ref.ABOUT')
        file_path = posixpath.join(posixpath.dirname(test_file), 'about_file_missing.c')
        a = model.About(test_file)
        err_msg = 'Field about_resource: Path %s not found' % file_path
        expected = [Error(INFO, err_msg)]
        result = a.errors
        assert expected == result

    def test_About_has_errors_when_missing_required_fields_are_missing(self):
        test_file = get_test_loc('test_model/parse/missing_required.ABOUT')
        a = model.About(test_file)
        expected = [
            Error(CRITICAL, 'Field about_resource is required'),
            Error(CRITICAL, 'Field name is required'),
        ]
        result = a.errors
        assert expected == result

    def test_About_has_errors_when_required_fields_are_empty(self):
        test_file = get_test_loc('test_model/parse/empty_required.ABOUT')
        a = model.About(test_file)
        expected = [
            Error(CRITICAL, 'Field about_resource is required and empty'),
            Error(CRITICAL, 'Field name is required and empty'),
        ]
        result = a.errors
        assert expected == result

    def test_About_has_errors_with_empty_notice_file_field(self):
        test_file = get_test_loc('test_model/parse/empty_notice_field.about')
        a = model.About(test_file)
        expected = [
            Error(INFO, 'Field notice_file is present but empty.')]
        result = a.errors
        assert expected == result

    def test_About_custom_fields_are_never_ignored(self):
        test_file = get_test_loc('test_model/custom_fields/custom_fields.about')
        a = model.About(test_file)
        result = [(n, f.value) for n, f in a.custom_fields.items()]
        expected = [
            (u'single_line', u'README STUFF'),
            (u'multi_line', u'line1\nline2'),
            (u'other', u'sasasas'),
            (u'empty', u'')
        ]

        assert expected == result

    def test_About_custom_fields_are_not_ignored_and_order_is_preserved(self):
        test_file = get_test_loc('test_model/custom_fields/custom_fields.about')
        a = model.About(test_file)
        result = [(n, f.value) for n, f in a.custom_fields.items()]
        expected = [
            (u'single_line', u'README STUFF'),
            (u'multi_line', u'line1\nline2'),
            (u'other', u'sasasas'),
            (u'empty', u'')
        ]
        assert sorted(expected) == sorted(result)

    def test_About_has_errors_for_illegal_custom_field_name(self):
        test_file = get_test_loc('test_model/parse/illegal_custom_field.about')
        a = model.About(test_file)
        expected_errors = [
            Error(INFO, 'Custom Field: hydrate'),
            Error(CRITICAL, "Internal error with custom field: 'hydrate': 'illegal name'.")
        ]

        assert expected_errors == a.errors
        assert not hasattr(getattr(a, 'hydrate'), 'value')
        field = list(a.custom_fields.values())[0]
        assert 'hydrate' == field.name
        assert 'illegal name' == field.value

    def test_About_file_fields_are_empty_if_present_and_path_missing(self):
        test_file = get_test_loc('test_model/parse/missing_notice_license_files.ABOUT')
        a = model.About(test_file)

        file_path1 = posixpath.join(posixpath.dirname(test_file), 'test.LICENSE')
        file_path2 = posixpath.join(posixpath.dirname(test_file), 'test.NOTICE')

        err_msg1 = Error(CRITICAL, 'Field license_file: Path %s not found' % file_path1)
        err_msg2 = Error(CRITICAL, 'Field notice_file: Path %s not found' % file_path2)

        expected_errors = [err_msg1, err_msg2]
        assert expected_errors == a.errors

        assert {'test.LICENSE': None} == a.license_file.value
        assert {'test.NOTICE': None} == a.notice_file.value

    def test_About_notice_and_license_text_are_loaded_from_file(self):
        test_file = get_test_loc('test_model/parse/license_file_notice_file.ABOUT')
        a = model.About(test_file)

        expected = '''Tester holds the copyright for test component. Tester relinquishes copyright of
this software and releases the component to Public Domain.

* Email Test@tester.com for any questions'''

        result = a.license_file.value['license_text.LICENSE']
        assert expected == result

        expected = '''Test component is released to Public Domain.'''
        result = a.notice_file.value['notice_text.NOTICE']
        assert expected == result

    def test_About_license_and_notice_text_are_empty_if_field_missing(self):
        test_file = get_test_loc('test_model/parse/no_file_fields.ABOUT')
        a = model.About(test_file)
        assert [] == a.errors
        assert {} == a.license_file.value
        assert {} == a.notice_file.value

    def test_About_rejects_non_ascii_names_and_accepts_unicode_values(self):
        test_file = get_test_loc('test_model/parse/non_ascii_field_name_value.about')
        a = model.About(test_file)
        expected = [
            Error(ERROR, "Field name: ['mat\xedas'] contains illegal name characters (or empty spaces) and is ignored.")
        ]
        assert expected == a.errors

    def test_About_invalid_boolean_value(self):
        test_file = get_test_loc('test_model/parse/invalid_boolean.about')
        a = model.About(test_file)
        expected_msg = "Field modified: Invalid flag value: 'blah'"
        assert expected_msg in a.errors[0].message

    def test_About_boolean_value(self):
        test_file = get_test_loc('test_model/parse/boolean_data.about')
        a = model.About(test_file)
        expected_msg = "Field track_changes is present but empty."
        assert expected_msg in a.errors[0].message
        # Context of the test file
        """
        about_resource: .
        name: boolean_data
        attribute: False
        modified: true
        internal_use_only: no
        redistribute: yes
        track_changes:
        """
        assert a.attribute.value is False
        assert a.modified.value is True
        assert a.internal_use_only.value is False
        assert a.redistribute.value is True
        assert a.track_changes.value is None

    def test_About_contains_about_file_path(self):
        test_file = get_test_loc('test_model/serialize/about.ABOUT')
        # TODO: I am not sure this override of the about_file_path makes sense
        a = model.About(test_file, about_file_path='complete/about.ABOUT')
        assert [] == a.errors
        expected = 'complete/about.ABOUT'
        result = a.about_file_path
        assert expected == result

    def test_About_equals(self):
        test_file = get_test_loc('test_model/equal/complete/about.ABOUT')
        a = model.About(test_file, about_file_path='complete/about.ABOUT')
        b = model.About(test_file, about_file_path='complete/about.ABOUT')
        assert a == b

    def test_About_are_not_equal_with_small_text_differences(self):
        test_file = get_test_loc('test_model/equal/complete2/about.ABOUT')
        a = model.About(test_file, about_file_path='complete2/about.ABOUT')
        test_file2 = get_test_loc('test_model/equal/complete/about.ABOUT')
        b = model.About(test_file2, about_file_path='complete/about.ABOUT')
        assert a.dumps() != b.dumps()
        assert a != b

    def test_get_field_names_only_returns_non_empties(self):
        a = model.About()
        a.custom_fields['f'] = model.StringField(
            name='f', value='1', present=True)
        b = model.About()
        b.custom_fields['g'] = model.StringField(
            name='g', value='1', present=True)
        abouts = [a, b]
        # ensure all fields (including custom fields) and
        # about_resource are collected in the correct order
        expected = [
            model.About.ABOUT_RESOURCE_ATTR, 'name', 'f', 'g'
        ]
        result = model.get_field_names(abouts)
        assert expected == result

    def test_get_field_names_does_not_return_duplicates_custom_fields(self):
        a = model.About()
        a.custom_fields['f'] = model.StringField(name='f', value='1',
                                                 present=True)
        a.custom_fields['cf'] = model.StringField(name='cf', value='1',
                                                 present=True)
        b = model.About()
        b.custom_fields['g'] = model.StringField(name='g', value='1',
                                                 present=True)
        b.custom_fields['cf'] = model.StringField(name='cf', value='2',
                                                 present=True)
        abouts = [a, b]
        # ensure all fields (including custom fields) and
        # about_resource are collected in the correct order
        expected = [
            'about_resource',
            'name',
            'cf',
            'f',
            'g',
            ]
        result = model.get_field_names(abouts)
        assert expected == result

    def test_comma_in_license(self):
        test_file = get_test_loc('test_model/special_char/about.ABOUT')
        a = model.About(test_file)
        expected = Error(ERROR, "The following character(s) cannot be in the license_key: [',']")
        assert a.errors[0] == expected

    def test_load_dict_issue_433(self):
        package_data = {
            'about_resource': 'package1.zip',
            'name': 'package',
            'version': '1.0',
            'copyright': 'copyright on package',
            'license_expression': 'license1 AND license2',
            'notice_file': 'package1.zip.NOTICE',
            'licenses': [
                {'key': 'license1', 'name': 'License1', 'file': 'license1.LICENSE', 'url': 'some_url', 'spdx_license_key': 'key'},
                {'key': 'license2', 'name': 'License2', 'file': 'license2.LICENSE', 'url': 'some_url', 'spdx_license_key': 'key'},
            ],
        }
        about = model.About()
        about.load_dict(package_data, base_dir='')
        as_dict = about.as_dict()
        expected = '''about_resource: package1.zip
name: package
version: '1.0'
license_expression: license1 AND license2
copyright: copyright on package
notice_file: package1.zip.NOTICE
licenses:
  - key: license1
    name: License1
    file: license1.LICENSE
    url: some_url
    spdx_license_key: key
  - key: license2
    name: License2
    file: license2.LICENSE
    url: some_url
    spdx_license_key: key
'''
        lic_dict = {u'license1': [u'License1', u'license1.LICENSE',u'', u'some_url', 'key'], u'license2' : [u'License2', u'license2.LICENSE', u'', u'some_url', 'key']}
        assert about.dumps(lic_dict) == expected


class SerializationTest(unittest.TestCase):

    def test_About_dumps(self):
        test_file = get_test_loc('test_model/dumps/about.ABOUT')
        a = model.About(test_file)
        assert [] == a.errors

        expected = '''about_resource: .
name: AboutCode
version: 0.11.0
description: |
  AboutCode is a tool
  to process ABOUT files.
  An ABOUT file is a file.
homepage_url: http://dejacode.org
license_expression: apache-2.0
copyright: Copyright (c) 2013-2014 nexB Inc.
notice_file: NOTICE
owner: nexB Inc.
author: Jillian Daguil, Chin Yeung Li, Philippe Ombredanne, Thomas Druez
vcs_tool: git
vcs_repository: https://github.com/dejacode/about-code-tool.git
licenses:
  - key: apache-2.0
    name: Apache 2.0
    file: apache-2.0.LICENSE
'''
        result = a.dumps()
        assert expected == result

    def test_About_dumps_does_all_non_empty_present_fields(self):
        test_file = get_test_loc('test_model/parse/complete2/about.ABOUT')
        a = model.About(test_file)
        expected_error = [
            Error(INFO, 'Custom Field: custom1'),
            Error(INFO, 'Custom Field: custom2'),
            Error(INFO, 'Field custom2 is present but empty.')
        ]
        assert sorted(expected_error) == sorted(a.errors)

        expected = '''about_resource: .
name: AboutCode
version: 0.11.0
custom1: |
  multi
  line
'''
        result = a.dumps()
        assert expected == result

    def test_About_dumps_with_different_boolean_value(self):
        test_file = get_test_loc('test_model/parse/complete2/about2.ABOUT')
        a = model.About(test_file)
        expected_error_msg = "Field track_changes: Invalid flag value: 'blah' is not one of"
        assert len(a.errors) == 1
        assert expected_error_msg in a.errors[0].message

        expected = '''about_resource: .

name: AboutCode
version: 0.11.0

redistribute: no
attribute: yes
modified: yes
'''

        result = a.dumps()
        assert set(expected) == set(result)

    def test_About_dumps_all_non_empty_fields(self):
        test_file = get_test_loc('test_model/parse/complete2/about.ABOUT')
        a = model.About(test_file)
        expected_error = [
            Error(INFO, 'Custom Field: custom1'),
            Error(INFO, 'Custom Field: custom2'),
            Error(INFO, 'Field custom2 is present but empty.')
        ]
        assert sorted(expected_error) == sorted(a.errors)

        expected = '''about_resource: .
name: AboutCode
version: 0.11.0
custom1: |
  multi
  line
'''
        result = a.dumps()
        assert expected == result

    def test_About_as_dict_contains_special_paths(self):
        test_file = get_test_loc('test_model/special/about.ABOUT')
        a = model.About(test_file, about_file_path='complete/about.ABOUT')
        expected_errors = []
        assert expected_errors == a.errors
        as_dict = a.as_dict()
        expected = 'complete/about.ABOUT'
        result = as_dict[model.About.ABOUT_FILE_PATH_ATTR]
        assert expected == result

    def test_load_dump_is_idempotent(self):
        test_file = get_test_loc('test_model/this.ABOUT')
        a = model.About()
        a.load(test_file)
        dumped_file = get_temp_file('that.ABOUT')
        a.dump(dumped_file)

        expected = get_unicode_content(test_file).splitlines()
        result = get_unicode_content(dumped_file).splitlines()
        # Ignore comment and empty line
        filtered_result = []
        for line in result:
            if not line.startswith('#') and not line == '':
                filtered_result.append(line)
        assert expected == filtered_result

    def test_load_can_load_unicode(self):
        test_file = get_test_loc('test_model/unicode/nose-selecttests.ABOUT')
        a = model.About()
        a.load(test_file)
        file_path = posixpath.join(posixpath.dirname(test_file), 'nose-selecttests-0.3.zip')
        err_msg = 'Field about_resource: Path %s not found' % file_path
        errors = [
            Error(INFO, 'Custom Field: dje_license'),
            Error(INFO, 'Custom Field: license_text_file'),
            Error(INFO, 'Custom Field: scm_tool'),
            Error(INFO, 'Custom Field: scm_repository'),
            Error(INFO, 'Custom Field: test'),
            Error(INFO, err_msg)]

        assert errors == a.errors
        assert 'Copyright (c) 2012, Domen Kožar' == a.copyright.value

    def test_load_non_unicode(self):
        test_file = get_test_loc('test_model/unicode/not-unicode.ABOUT')
        a = model.About()
        a.load(test_file)
        err = a.errors[0]
        assert CRITICAL == err.severity
        assert 'Cannot load invalid ABOUT file' in err.message

    def test_as_dict_load_dict_ignores_empties(self):
        test = {
            'about_resource': '.',
            'author': '',
            'copyright': 'Copyright (c) 2013-2014 nexB Inc.',
            'custom1': 'some custom',
            'custom_empty': '',
            'description': 'AboutCode is a tool\nfor files.',
            'license_expression': 'apache-2.0',
            'name': 'AboutCode',
            'owner': 'nexB Inc.'}

        expected = {
            'about_file_path': None,
            'about_resource': dict([('.', None)]),
            'copyright': 'Copyright (c) 2013-2014 nexB Inc.',
            'custom1': 'some custom',
            'description': 'AboutCode is a tool\nfor files.',
            'license_expression': 'apache-2.0',
            'name': 'AboutCode',
            'owner': 'nexB Inc.'}

        a = model.About()
        base_dir = 'some_dir'
        a.load_dict(test, base_dir)
        as_dict = a.as_dict()
        # FIXME: why converting back to dict?
        assert expected == dict(as_dict)

    def test_load_dict_as_dict_is_idempotent_ignoring_special(self):
        test = {
            'about_resource': ['.'],
            'attribute': 'yes',
            'author': 'Jillian Daguil, Chin Yeung Li, Philippe Ombredanne, Thomas Druez',
            'copyright': 'Copyright (c) 2013-2014 nexB Inc.',
            'description': 'AboutCode is a tool to process ABOUT files. An ABOUT file is a file.',
            'homepage_url': 'http://dejacode.org',
            'license_expression': 'apache-2.0',
            'name': 'AboutCode',
            'owner': 'nexB Inc.',
            'vcs_repository': 'https://github.com/dejacode/about-code-tool.git',
            'vcs_tool': 'git',
            'version': '0.11.0'}
        a = model.About()
        base_dir = 'some_dir'
        a.load_dict(test, base_dir)
        as_dict = a.as_dict()

        expected = {
            'about_file_path': None,
            'about_resource': dict([('.', None)]),
            'attribute': 'yes',
            'author': 'Jillian Daguil, Chin Yeung Li, Philippe Ombredanne, Thomas Druez',
            'copyright': 'Copyright (c) 2013-2014 nexB Inc.',
            'description': 'AboutCode is a tool to process ABOUT files. An ABOUT file is a file.',
            'homepage_url': 'http://dejacode.org',
            'license_expression': 'apache-2.0',
            'name': 'AboutCode',
            'owner': 'nexB Inc.',
            'vcs_repository': 'https://github.com/dejacode/about-code-tool.git',
            'vcs_tool': 'git',
            'version': '0.11.0'}

        assert expected == dict(as_dict)

    def test_about_model_class_from_dict_constructor(self):
        about_data = {
            'about_resource': ['.'],
            'attribute': 'yes',
            'author': 'Jillian Daguil, Chin Yeung Li, Philippe Ombredanne, Thomas Druez',
            'copyright': 'Copyright (c) 2013-2014 nexB Inc.',
            'description': 'AboutCode is a tool to process ABOUT files. An ABOUT file is a file.',
            'homepage_url': 'http://dejacode.org',
            'license_expression': 'apache-2.0',
            'name': 'AboutCode',
            'owner': 'nexB Inc.',
            'vcs_repository': 'https://github.com/dejacode/about-code-tool.git',
            'vcs_tool': 'git',
            'version': '0.11.0',
        }

        about = model.About.from_dict(about_data)
        assert isinstance(about, model.About)

        about_data.update({
            'about_file_path': None,
            'about_resource': dict([('.', None)]),
        })
        assert about_data == about.as_dict()

    def test_write_output_csv(self):
        path = 'test_model/this.ABOUT'
        test_file = get_test_loc(path)
        abouts = model.About(location=test_file, about_file_path=path)

        result = get_temp_file()
        model.write_output([abouts], result, format='csv')

        expected = get_test_loc('test_model/expected.csv')
        check_csv(expected, result)

    def test_write_output_csv_with_multiple_files(self):
        path = 'test_model/multiple_files.ABOUT'
        test_file = get_test_loc(path)
        abouts = model.About(location=test_file, about_file_path=path)

        result = get_temp_file()
        model.write_output([abouts], result, format='csv')

        expected = get_test_loc('test_model/multiple_files_expected.csv')
        check_csv(expected, result)

    def test_write_output_json(self):
        path = 'test_model/this.ABOUT'
        test_file = get_test_loc(path)
        abouts = model.About(location=test_file, about_file_path=path)

        result = get_temp_file()
        model.write_output([abouts], result, format='json')

        expected = get_test_loc('test_model/expected.json')
        check_json(expected, result)

    def test_android_module_license(self):
        path = 'test_model/android/single_license.c.ABOUT'
        test_file = get_test_loc(path)
        abouts = model.About(location=test_file, about_file_path=path)

        parent_dir = get_temp_dir()
        abouts.android_module_license(parent_dir)
        assert os.path.exists(os.path.join(parent_dir, 'MODULE_LICENSE_PUBLIC_DOMAIN'))

    def test_android_module_multi_licenses(self):
        path = 'test_model/android/multi_license.c.ABOUT'
        test_file = get_test_loc(path)
        abouts = model.About(location=test_file, about_file_path=path)

        parent_dir = get_temp_dir()
        abouts.android_module_license(parent_dir)
        assert os.path.exists(os.path.join(parent_dir, 'MODULE_LICENSE_BSD_NEW'))
        assert os.path.exists(os.path.join(parent_dir, 'MODULE_LICENSE_BSD_SIMPLIFIED'))

    def test_android_notice(self):
        path = 'test_model/android/single_license.c.ABOUT'
        test_file = get_test_loc(path)
        abouts = model.About(location=test_file, about_file_path=path)

        parent_dir = get_temp_dir()
        notice_path, notice_context = abouts.android_notice(parent_dir)
        expected_path = os.path.join(parent_dir, 'NOTICE')
        assert os.path.normpath(notice_path) == expected_path

        expected_notice = '''Copyright (c) xyz

This component is released to the public domain by the author.

'''
        assert notice_context == expected_notice


class CollectorTest(unittest.TestCase):

    def test_collect_inventory_return_errors(self):
        test_loc = get_test_loc('test_model/collect_inventory_errors')
        errors, _abouts = model.collect_inventory(test_loc)
        file_path1 = posixpath.join(test_loc, 'distribute_setup.py')
        file_path2 = posixpath.join(test_loc, 'date_test.py')

        err_msg1 = 'non-supported_date_format.ABOUT: Field about_resource: Path %s not found' % file_path1
        err_msg2 = 'supported_date_format.ABOUT: Field about_resource: Path %s not found' % file_path2
        expected_errors = [
            Error(INFO, "Field ['date'] is a custom field."),
            Error(INFO, err_msg1),
            Error(INFO, err_msg2)]
        assert sorted(expected_errors) == sorted(errors)

    def test_collect_inventory_with_long_path(self):
        test_loc = extract_test_loc('test_model/longpath.zip')
        _errors, abouts = model.collect_inventory(test_loc)
        assert 2 == len(abouts)

        expected_paths = (
            'longpath/longpath1/longpath1/longpath1/longpath1/longpath1/longpath1'
            '/longpath1/longpath1/longpath1/longpath1/longpath1/longpath1/longpath1'
            '/longpath1/longpath1/longpath1/longpath1/longpath1/longpath1/longpath1'
            '/longpath1/longpath1/longpath1/longpath1/longpath1/longpath1/longpath1'
            '/longpath1/non-supported_date_format.ABOUT',
            'longpath/longpath1/longpath1/longpath1/longpath1/longpath1/longpath1'
            '/longpath1/longpath1/longpath1/longpath1/longpath1/longpath1/longpath1'
            '/longpath1/longpath1/longpath1/longpath1/longpath1/longpath1/longpath1'
            '/longpath1/longpath1/longpath1/longpath1/longpath1/longpath1/longpath1'
            '/longpath1/supported_date_format.ABOUT'
        )
        results = [a.about_file_path for a in abouts]
        assert all(r.endswith(expected_paths) for r in results)

        expected_name = ['distribute', 'date_test']
        result_name = [a.name.value for a in abouts]
        assert sorted(expected_name) == sorted(result_name)

    def test_collect_inventory_can_collect_a_single_file(self):
        test_loc = get_test_loc('test_model/single_file/django_snippets_2413.ABOUT')
        _errors, abouts = model.collect_inventory(test_loc)
        assert 1 == len(abouts)
        expected = ['single_file/django_snippets_2413.ABOUT']
        result = [a.about_file_path for a in abouts]
        assert expected == result

    def test_collect_inventory_return_no_warnings_and_model_can_use_relative_paths(self):
        test_loc = get_test_loc('test_model/rel/allAboutInOneDir')
        errors, _abouts = model.collect_inventory(test_loc)
        expected_errors = []
        result = [(level, e) for level, e in errors if level > INFO]
        assert expected_errors == result

    def test_collect_inventory_populate_about_file_path(self):
        test_loc = get_test_loc('test_model/inventory/complete')
        errors, abouts = model.collect_inventory(test_loc)
        assert [] == errors
        expected = 'about.ABOUT'
        result = abouts[0].about_file_path
        assert expected == result

    def test_collect_inventory_with_multi_line(self):
        test_loc = get_test_loc('test_model/parse/multi_line_license_expresion.ABOUT')
        errors, abouts = model.collect_inventory(test_loc)
        assert [] == errors
        expected_lic_url = [
            'https://enterprise.dejacode.com/urn/?urn=urn:dje:license:mit',
            'https://enterprise.dejacode.com/urn/?urn=urn:dje:license:apache-2.0']
        returned_lic_url = abouts[0].license_url.value
        assert expected_lic_url == returned_lic_url

    def test_collect_inventory_with_license_expression(self):
        test_loc = get_test_loc('test_model/parse/multi_line_license_expresion.ABOUT')
        errors, abouts = model.collect_inventory(test_loc)
        assert [] == errors
        expected_lic = 'mit or apache-2.0'
        returned_lic = abouts[0].license_expression.value
        assert expected_lic == returned_lic

    def test_collect_inventory_always_collects_custom_fieldsg(self):
        test_loc = get_test_loc('test_model/inventory/custom_fields.ABOUT')
        errors, abouts = model.collect_inventory(test_loc)
        expected_msg = "Field ['resource', 'custom_mapping'] is a custom field."
        assert len(errors) == 1
        assert expected_msg in errors[0].message
        # The value of the custom field: 'resource' is collected
        assert abouts[0].resource.value

    def test_collect_inventory_does_not_raise_error_and_maintains_order_on_custom_fields(self):
        test_loc = get_test_loc('test_model/inventory/custom_fields2.ABOUT')
        errors, abouts = model.collect_inventory(test_loc)
        expected_errors = [
            Error(INFO, "Field ['resource', 'custom_mapping'] is a custom field.")
        ]
        assert expected_errors == errors
        expected = [u'about_resource: .\nname: test\nresource: .\ncustom_mapping: test\n']
        assert expected == [a.dumps() for a in abouts]

    def test_parse_license_expression(self):
        spec_char, returned_lic = model.parse_license_expression('mit or apache-2.0')
        expected_lic = ['mit', 'apache-2.0']
        expected_spec_char = []
        assert expected_lic == returned_lic
        assert expected_spec_char == spec_char

    def test_parse_license_expression_with_special_chara(self):
        spec_char, returned_lic = model.parse_license_expression('mit, apache-2.0')
        expected_lic = []
        expected_spec_char = [',']
        assert expected_lic == returned_lic
        assert expected_spec_char == spec_char

    def test_collect_inventory_works_with_relative_paths(self):
        # FIXME: This test need to be run under src/attributecode/
        # or otherwise it will fail as the test depends on the launching
        # location
        test_loc = get_test_loc('test_model/inventory/relative')
        # Use '.' as the indication of the current directory
        test_loc1 = test_loc + '/./'
        # Use '..' to go back to the parent directory
        test_loc2 = test_loc + '/../relative'
        errors1, abouts1 = model.collect_inventory(test_loc1)
        errors2, abouts2 = model.collect_inventory(test_loc2)
        assert [] == errors1
        assert [] == errors2
        expected = 'about.ABOUT'
        result1 = abouts1[0].about_file_path
        result2 = abouts2[0].about_file_path
        assert expected == result1
        assert expected == result2

    def test_collect_inventory_basic_from_directory(self):
        location = get_test_loc('test_model/inventory/basic')
        result = get_temp_file()
        errors, abouts = model.collect_inventory(location)

        model.write_output(abouts, result, format='csv')

        expected_errors = []
        assert expected_errors == errors

        expected = get_test_loc('test_model/inventory/basic/expected.csv')
        check_csv(expected, result)

    def test_collect_inventory_with_about_resource_path_from_directory(self):
        location = get_test_loc('test_model/inventory/basic_with_about_resource_path')
        result = get_temp_file()
        errors, abouts = model.collect_inventory(location)

        model.write_output(abouts, result, format='csv')

        expected_errors = []
        assert expected_errors == errors

        expected = get_test_loc('test_model/inventory/basic_with_about_resource_path/expected.csv')
        check_csv(expected, result)

    def test_collect_inventory_with_no_about_resource_from_directory(self):
        location = get_test_loc('test_model/inventory/no_about_resource_key')
        result = get_temp_file()
        errors, abouts = model.collect_inventory(location)

        model.write_output(abouts, result, format='csv')

        expected_errors = [Error(CRITICAL, 'about/about.ABOUT: Field about_resource is required')]
        assert expected_errors == errors

    def test_collect_inventory_complex_from_directory(self):
        location = get_test_loc('test_model/inventory/complex')
        result = get_temp_file()
        errors, abouts = model.collect_inventory(location)

        model.write_output(abouts, result, format='csv')

        assert all(e.severity == INFO for e in errors)

        expected = get_test_loc('test_model/inventory/complex/expected.csv')
        check_csv(expected, result, fix_cell_linesep=True, regen=False)

    def test_collect_inventory_does_not_convert_lf_to_crlf_from_directory(self):
        location = get_test_loc('test_model/crlf/about.ABOUT')
        result = get_temp_file()
        errors, abouts = model.collect_inventory(location)
        model.write_output(abouts, result, format='csv')
        assert all(e.severity == INFO for e in errors)

        expected = get_test_loc('test_model/crlf/expected.csv')
        check_csv(expected, result, fix_cell_linesep=True, regen=False)

    def test_copy_redist_src_no_structure(self):
        test_loc = get_test_loc('test_model/redistribution/')
        copy_list = [get_test_loc('test_model/redistribution/this.c'), get_test_loc('test_model/redistribution/test/subdir')]
        output = get_temp_dir()

        expected_file = ['this.c', 'subdir']

        with_structure = False
        err = model.copy_redist_src(copy_list, test_loc, output, with_structure)

        assert err == []

        from os import listdir
        copied_files = listdir(output)
        assert len(expected_file) == len(copied_files)
        assert err == []
        for file in expected_file:
            assert file in copied_files

    def test_copy_redist_src_with_structure(self):
        test_loc = get_test_loc('test_model/redistribution/')
        copy_list = [get_test_loc('test_model/redistribution/this.c'), get_test_loc('test_model/redistribution/test/subdir')]
        output = get_temp_dir()

        expected_file = ['this.c', 'test']

        with_structure = True
        err = model.copy_redist_src(copy_list, test_loc, output, with_structure)

        assert err == []

        from os import listdir
        copied_files = listdir(output)
        assert len(expected_file) == len(copied_files)
        assert err == []
        for file in expected_file:
            assert file in copied_files

    def test_get_copy_list(self):
        location = get_test_loc('test_model/redistribution/')
        result = get_temp_file()
        errors, abouts = model.collect_inventory(location)
        copy_list, err = model.get_copy_list(abouts, location)
        assert err == []
        expected = [os.path.join(location, 'this.c'), os.path.join(location, 'test/subdir')]
        if on_windows:
            norm_list = []
            for c in copy_list:
                norm_list.append(norm(c))
            assert norm_list == expected
        else:
            assert copy_list == expected


class FetchLicenseTest(unittest.TestCase):

    @mock.patch.object(model, 'urlopen')
    def test_valid_api_url(self, mock_data):
        mock_data.return_value = ''
        assert model.valid_api_url('non_valid_url') is False

    @mock.patch('attributecode.util.have_network_connection')
    @mock.patch('attributecode.model.valid_api_url')
    def test_pre_process_and_fetch_license_dict_dje(self, have_network_connection, valid_api_url):
        have_network_connection.return_value = True
        valid_api_url.return_value = False
        error_msg = (
            'Network problem. Please check your Internet connection. '
            'License generation is skipped.')
        expected = ({}, [Error(ERROR, error_msg)])
        assert model.pre_process_and_fetch_license_dict([]) == expected

        valid_api_url.return_value = True
        expected = ({}, [])
        assert model.pre_process_and_fetch_license_dict([]) == expected

    @mock.patch('attributecode.util.have_network_connection')
    @mock.patch('attributecode.model.valid_api_url')
    def test_pre_process_and_fetch_license_dict_licensedb(self, have_network_connection, valid_api_url):
        have_network_connection.return_value = False
        valid_api_url.return_value = False
        error_msg = (
            'Network problem. Please check your Internet connection. '
            'License generation is skipped.')
        expected = ({}, [Error(ERROR, error_msg)])
        assert model.pre_process_and_fetch_license_dict([]) == expected

        have_network_connection.return_value = True
        valid_api_url.return_value = True
        expected = ({}, [])

        assert model.pre_process_and_fetch_license_dict([]) == expected


#!/usr/bin/env python
# -*- coding: utf8 -*-

# ============================================================================
#  Copyright (c) nexB Inc. http://www.nexb.com/ - All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ============================================================================

import codecs

from posixpath import basename
from posixpath import dirname
from posixpath import exists
from posixpath import join
from posixpath import normpath

from attributecode import ERROR
from attributecode import CRITICAL
from attributecode import INFO
from attributecode import Error
from attributecode import model
from attributecode import util
from attributecode.util import add_unc
from attributecode.util import csv
from attributecode.util import file_fields
from attributecode.util import invalid_chars
from attributecode.util import to_posix
from attributecode.util import UNC_PREFIX_POSIX
from attributecode.util import unique
from attributecode.util import load_scancode_json, load_csv, load_json, load_excel


def check_duplicated_columns(location):
    """
    Return a list of errors for duplicated column names in a CSV file
    at location.
    """
    location = add_unc(location)
    with open(location, mode='r', encoding='utf-8-sig', errors='replace') as csvfile:
        reader = csv.reader(csvfile)
        columns = next(reader)
        columns = [col for col in columns]

    seen = set()
    dupes = dict()
    for col in columns:
        c = col.lower()
        if c in seen:
            if c in dupes:
                dupes[c].append(col)
            else:
                dupes[c] = [col]
        seen.add(c.lower())

    errors = []
    if dupes:
        dup_msg = []
        for name, names in dupes.items():
            names = u', '.join(names)
            msg = '%(name)s with %(names)s' % locals()
            dup_msg.append(msg)
        dup_msg = u', '.join(dup_msg)
        msg = ('Duplicated column name(s): %(dup_msg)s\n' % locals() +
               'Please correct the input and re-run.')
        err = Error(ERROR, msg)
        if not err in errors:
            errors.append(err)
    return errors


def check_duplicated_about_resource(arp, arp_list):
    """
    Return error for duplicated about_resource.
    """
    if arp in arp_list:
        msg = ("The input has duplicated values in 'about_resource' "
               "field: " + arp)
        return Error(CRITICAL, msg)
    return ''


def check_newline_in_file_field(component):
    """
    Return a list of errors for newline characters detected in *_file fields.
    """
    errors = []
    for k in component.keys():
        if k in file_fields:
            try:
                if '\n' in component[k]:
                    msg = ("New line character detected in '%s' for '%s' which is not supported."
                            "\nPlease use ',' to declare multiple files.") % (k, component['about_resource'])
                    errors.append(Error(CRITICAL, msg))
            except:
                pass
    return errors


def check_about_resource_filename(arp):
    """
    Return error for invalid/non-support about_resource's filename or
    empty string if no error is found.
    """
    if invalid_chars(arp):
        msg = ("Invalid characters present in 'about_resource' "
                   "field: " + arp)
        return (Error(ERROR, msg))
    return ''


def load_inventory(location, from_attrib=False, base_dir=None, scancode=False, reference_dir=None):
    """
    Load the inventory file at `location` for ABOUT and LICENSE files stored in
    the `base_dir`. Return a list of errors and a list of About objects
    validated against the `base_dir`.

    Optionally use `reference_dir` as the directory location of extra reference
    license and notice files to reuse.
    """
    errors = []
    abouts = []

    if base_dir:
        base_dir = util.to_posix(base_dir)
    if scancode:
        inventory = load_scancode_json(location)
    else:
        if location.endswith('.csv'):
            dup_cols_err = check_duplicated_columns(location)
            if dup_cols_err:
                errors.extend(dup_cols_err)
                return errors, abouts
            inventory = load_csv(location)
        elif location.endswith('.xlsx'):
            dup_cols_err, inventory = load_excel(location)
            if dup_cols_err:
                errors.extend(dup_cols_err)
                return errors, abouts
        else:
            inventory = load_json(location)

    try:
        arp_list = []
        errors = []
        for component in inventory:
            if not from_attrib:
                arp = component['about_resource']
                dup_err = check_duplicated_about_resource(arp, arp_list)
                if dup_err:
                    if not dup_err in errors:
                        errors.append(dup_err)
                else:
                    arp_list.append(arp)

                invalid_about_filename = check_about_resource_filename(arp)
                if invalid_about_filename and not invalid_about_filename in errors:
                    errors.append(invalid_about_filename)

            newline_in_file_err = check_newline_in_file_field(component)
            if newline_in_file_err:
                errors.extend(newline_in_file_err)
        if errors:
            return errors, abouts
    except Exception as e:
        # TODO: why catch ALL Exception
        msg = "The essential field 'about_resource' is not found in the <input>"
        errors.append(Error(CRITICAL, msg))
        return errors, abouts

    custom_fields_list = []
    for fields in inventory:
        # check does the input contains the required fields
        required_fields = model.About.required_fields

        for f in required_fields:
            if f not in fields:
                if from_attrib and f == 'about_resource':
                    continue
                else:
                    msg = "Required field: %(f)r not found in the <input>" % locals()
                    errors.append(Error(CRITICAL, msg))
                    return errors, abouts
        # Set about file path to '' if no 'about_resource' is provided from
        # the input for `attrib`
        if not 'about_resource' in fields:
            afp = ''
        else:
            afp = fields.get(model.About.ABOUT_RESOURCE_ATTR)

        """
        # FIXME: this should not be a failure condition
        if not afp or not afp.strip():
            msg = 'Empty column: %(afp)r. Cannot generate .ABOUT file.' % locals()
            errors.append(Error(ERROR, msg))
            continue
        else:
        """
        afp = util.to_posix(afp)
        if base_dir:
            loc = join(base_dir, afp)
        else:
            loc = afp
        about = model.About(about_file_path=afp)
        about.location = loc

        # Update value for 'about_resource'
        # keep only the filename or '.' if it's a directory
        if 'about_resource' in fields:
            updated_resource_value = u''
            resource_path = fields['about_resource']
            if resource_path.endswith(u'/'):
                updated_resource_value = u'.'
            else:
                updated_resource_value = basename(resource_path)
            fields['about_resource'] = updated_resource_value

        # Set 'about_resource' to '.' if no 'about_resource' is provided from
        # the input for `attrib`
        elif not 'about_resource' in fields and from_attrib:
            fields['about_resource'] = u'.'

        ld_errors = about.load_dict(
            fields,
            base_dir,
            scancode=scancode,
            from_attrib=from_attrib,
            running_inventory=False,
            reference_dir=reference_dir,
        )
        for severity, message in ld_errors:
            if 'Custom Field' in message:
                field_name = message.replace('Custom Field: ', '').strip()
                if not field_name in custom_fields_list:
                    custom_fields_list.append(field_name)
            else:
                errors.append(Error(severity, message))

        abouts.append(about)
    if custom_fields_list:
        custom_fields_err_msg = 'Field ' + str(custom_fields_list) + ' is a custom field.'
        errors.append(Error(INFO, custom_fields_err_msg))
    # Covert the license_score value from string to list of int
    # The licesne_score is not in the spec but is specify in the scancode license scan.
    # This key will be treated as a custom string field. Therefore, we need to
    # convert back to the list with float type for score.
    if scancode:
        for about in abouts:
            try:
                score_list = list(map(float, about.license_score.value.replace('[', '').replace(']', '').split(',')))
                about.license_score.value = score_list
            except:
                pass

    return errors, abouts


def update_about_resource(self):
    pass


def generate(location, base_dir, android=None, reference_dir=None, fetch_license=False, fetch_license_djc=False):
    """
    Load ABOUT data from a CSV inventory at `location`. Write ABOUT files to
    base_dir. Return errors and about objects.
    """
    not_exist_errors = []
    notice_dict = {}
    api_url = ''
    api_key = ''
    gen_license = False
    # FIXME: use two different arguments: key and url
    # Check if the fetch_license contains valid argument
    if fetch_license_djc:
        # Strip the ' and " for api_url, and api_key from input
        api_url = fetch_license_djc[0].strip("'").strip('"')
        api_key = fetch_license_djc[1].strip("'").strip('"')
        gen_license = True

    if fetch_license:
        gen_license = True

    # TODO: WHY use posix??
    bdir = to_posix(base_dir)

    errors, abouts = load_inventory(
        location=location,
        base_dir=bdir,
        reference_dir=reference_dir
    )

    if gen_license:
        license_dict, err = model.pre_process_and_fetch_license_dict(abouts, api_url, api_key)
        if err:
            for e in err:
                # Avoid having same error multiple times
                if not e in errors:
                    errors.append(e)

    for about in abouts:
        # Strip trailing spaces
        about.about_file_path = about.about_file_path.strip()
        if about.about_file_path.startswith('/'):
            about.about_file_path = about.about_file_path.lstrip('/')
        dump_loc = join(bdir, about.about_file_path.lstrip('/'))

        # The following code is to check if there is any directory ends with spaces
        split_path = about.about_file_path.split('/')
        dir_endswith_space = False
        for segment in split_path:
            if segment.endswith(' '):
                msg = (u'File path : '
                       u'%(dump_loc)s '
                       u'contains directory name ends with spaces which is not '
                       u'allowed. Generation skipped.' % locals())
                errors.append(Error(ERROR, msg))
                dir_endswith_space = True
                break
        if dir_endswith_space:
            # Continue to work on the next about object
            continue

        try:
            # Generate value for 'about_resource' if it does not exist
            if not about.about_resource.value:
                about.about_resource.value = dict()
                about_resource_value = ''
                if about.about_file_path.endswith('/'):
                    about_resource_value = u'.'
                else:
                    about_resource_value = basename(about.about_file_path)
                about.about_resource.value[about_resource_value] = None
                about.about_resource.present = True
                # Check for the existence of the 'about_resource'
                # If the input already have the 'about_resource' field, it will
                # be validated when creating the about object
                loc = util.to_posix(dump_loc)
                about_file_loc = loc
                path = join(dirname(util.to_posix(about_file_loc)), about_resource_value)
                if not exists(path):
                    path = util.to_posix(path.strip(UNC_PREFIX_POSIX))
                    path = normpath(path)
                    msg = (u'Field about_resource: '
                           u'%(path)s '
                           u'does not exist' % locals())
                errors.append(Error(INFO, msg))

            licenses_dict = {}
            if gen_license:
                # Write generated LICENSE file
                license_key_name_context_url_list = about.dump_lic(dump_loc, license_dict)
                if license_key_name_context_url_list:
                    for lic_key, lic_name, lic_filename, lic_context, lic_url, spdx_lic_key in license_key_name_context_url_list:
                        licenses_dict[lic_key] = [lic_name, lic_filename, lic_context, lic_url, spdx_lic_key]
                        if not lic_name in about.license_name.value:
                            about.license_name.value.append(lic_name)
                        about.license_file.value[lic_filename] = lic_filename
                        if not lic_url in about.license_url.value:
                            about.license_url.value.append(lic_url)
                        if not spdx_lic_key in about.spdx_license_key.value:
                            about.spdx_license_key.value.append(spdx_lic_key)
                        if about.license_name.value:
                            about.license_name.present = True
                        if about.license_file.value:
                            about.license_file.present = True
                        if about.license_url.value:
                            about.license_url.present = True
                        if about.spdx_license_key.value:
                            about.spdx_license_key.present = True

            about.dump(dump_loc, licenses_dict)

            if android:
                """
                Create MODULE_LICENSE_XXX and get context to create NOTICE file
                follow the standard from Android Open Source Project
                """
                import os
                parent_path = os.path.dirname(util.to_posix(dump_loc))

                about.android_module_license(parent_path)
                notice_path, notice_context = about.android_notice(parent_path)
                if notice_path in notice_dict.keys():
                    notice_dict[notice_path] += '\n\n' + notice_context
                else:
                    notice_dict[notice_path] = notice_context

        except Exception as e:
            # only keep the first 100 char of the exception
            # TODO: truncated errors are likely making diagnotics harder
            emsg = repr(e)[:100]
            msg = (u'Failed to write .ABOUT file at : '
                   u'%(dump_loc)s '
                   u'with error: %(emsg)s' % locals())
            errors.append(Error(ERROR, msg))

    if android:
        # Check if there is already a NOTICE file present
        for path in notice_dict.keys():
            if os.path.exists(path):
                msg = (u'NOTICE file already exist at: %s' % path)
                errors.append(Error(ERROR, msg))
            else:
                about.dump_android_notice(path, notice_dict[path])
    return errors, abouts



#!/usr/bin/env python
# -*- coding: utf8 -*-

# ============================================================================
#  Copyright (c) nexB Inc. http://www.nexb.com/ - All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ============================================================================

import io
import unittest

from attributecode import CRITICAL
from attributecode import DEBUG
from attributecode import ERROR
from attributecode import INFO
from attributecode import NOTSET
from attributecode import WARNING
from attributecode import cmd
from attributecode import Error

from testing_utils import run_about_command_test_click
from testing_utils import get_test_loc
from testing_utils import get_temp_dir
from testing_utils import get_temp_file

# NB: the test_report_errors* tests depend on py.test stdout/err capture capabilities


def test_report_errors(capsys):
    errors = [
        Error(CRITICAL, 'msg1'),
        Error(ERROR, 'msg2'),
        Error(INFO, 'msg3'),
        Error(WARNING, 'msg4'),
        Error(DEBUG, 'msg4'),
        Error(NOTSET, 'msg4'),
    ]
    ec = cmd.report_errors(errors, quiet=False, verbose=True, log_file_loc=None)
    assert 6 == ec
    out, err = capsys.readouterr()
    expected_out = [
        'Command completed with 6 errors or warnings.',
        'CRITICAL: msg1',
        'ERROR: msg2',
        'INFO: msg3',
        'WARNING: msg4',
        'DEBUG: msg4',
        'NOTSET: msg4']
    assert '' == err
    assert expected_out == out.splitlines(False)


def test_report_errors_without_verbose(capsys):
    errors = [
        Error(CRITICAL, 'msg1'),
        Error(ERROR, 'msg2'),
        Error(INFO, 'msg3'),
        Error(WARNING, 'msg4'),
        Error(DEBUG, 'msg4'),
        Error(NOTSET, 'msg4'),
    ]
    ec = cmd.report_errors(errors, quiet=False, verbose=False, log_file_loc=None)
    assert 3 == ec
    out, err = capsys.readouterr()
    expected_out = [
        'Command completed with 3 errors or warnings.',
        'CRITICAL: msg1',
        'ERROR: msg2',
        'WARNING: msg4',
    ]
    assert '' == err
    assert expected_out == out.splitlines(False)


def test_report_errors_with_quiet_ignores_verbose_flag(capsys):
    errors = [
        Error(CRITICAL, 'msg1'),
        Error(ERROR, 'msg2'),
        Error(INFO, 'msg3'),
        Error(WARNING, 'msg4'),
        Error(DEBUG, 'msg4'),
        Error(NOTSET, 'msg4'),
        Error(WARNING, 'msg4'),
    ]
    severe_errors_count = cmd.report_errors(errors, quiet=True, verbose=True)
    assert severe_errors_count == 6
    out, err = capsys.readouterr()
    assert '' == out
    assert '' == err


def test_report_errors_with_quiet_ignores_verbose_flag2(capsys):
    errors = [
        Error(CRITICAL, 'msg1'),
        Error(ERROR, 'msg2'),
        Error(INFO, 'msg3'),
        Error(WARNING, 'msg4'),
        Error(DEBUG, 'msg4'),
        Error(NOTSET, 'msg4'),
        Error(WARNING, 'msg4'),
    ]
    severe_errors_count = cmd.report_errors(errors, quiet=True, verbose=False)
    assert severe_errors_count == 3
    out, err = capsys.readouterr()
    assert '' == out
    assert '' == err


def test_report_errors_with_verbose_flag(capsys):
    errors = [
        Error(CRITICAL, 'msg1'),
        Error(ERROR, 'msg2'),
        Error(INFO, 'msg3'),
        Error(WARNING, 'msg4'),
        Error(DEBUG, 'msg4'),
        Error(NOTSET, 'msg4'),
        Error(WARNING, 'msg4'),
    ]
    severe_errors_count = cmd.report_errors(errors, quiet=False, verbose=True)
    assert severe_errors_count == 6
    out, err = capsys.readouterr()
    expected_out = [
        'Command completed with 6 errors or warnings.',
        'CRITICAL: msg1',
        'ERROR: msg2',
        'INFO: msg3',
        'WARNING: msg4',
        'DEBUG: msg4',
        'NOTSET: msg4'
    ]
    assert expected_out == out.splitlines(False)
    assert '' == err


def test_report_errors_can_write_to_logfile():
    errors = [
        Error(CRITICAL, 'msg1'),
        Error(ERROR, 'msg2'),
        Error(INFO, 'msg3'),
        Error(WARNING, 'msg4'),
        Error(DEBUG, 'msg4'),
        Error(NOTSET, 'msg4'),
        Error(WARNING, 'msg4'),
    ]

    result_file = get_temp_file()
    _ec = cmd.report_errors(errors, quiet=False, verbose=True,
                           log_file_loc=result_file)
    with open(result_file, 'r', encoding='utf-8', errors='replace') as rf:
        result = rf.read()
    expected = [
        'Command completed with 6 errors or warnings.',
        'CRITICAL: msg1',
        'ERROR: msg2',
        'INFO: msg3',
        'WARNING: msg4',
        'DEBUG: msg4',
        'NOTSET: msg4'
    ]
    assert expected == result.splitlines(False)


def test_report_errors_does_not_report_duplicate_errors(capsys):
    errors = [
        Error(CRITICAL, 'msg1'),
        Error(ERROR, 'msg2'),
        Error(INFO, 'msg3'),
        Error(WARNING, 'msg4'),
        Error(DEBUG, 'msg4'),
        Error(NOTSET, 'msg4'),
        # dupes
        Error(WARNING, 'msg4'),
        Error(CRITICAL, 'msg1'),
    ]
    severe_errors_count = cmd.report_errors(errors, quiet=True, verbose=True)
    assert severe_errors_count == 6


def test_get_error_messages():
    errors = [
        Error(CRITICAL, 'msg1'),
        Error(ERROR, 'msg2'),
        Error(INFO, 'msg3'),
        Error(WARNING, 'msg4'),
        Error(DEBUG, 'msg4'),
        Error(NOTSET, 'msg4'),
    ]

    emsgs, ec = cmd.get_error_messages(errors)
    assert 3 == ec
    expected = [
        'Command completed with 3 errors or warnings.',
        'CRITICAL: msg1',
        'ERROR: msg2',
        'WARNING: msg4',
    ]
    assert expected == emsgs


def test_get_error_messages_verbose():
    errors = [
        Error(CRITICAL, 'msg1'),
        Error(ERROR, 'msg2'),
        Error(INFO, 'msg3'),
        Error(WARNING, 'msg4'),
        Error(DEBUG, 'msg4'),
        Error(NOTSET, 'msg4'),
    ]

    emsgs, ec = cmd.get_error_messages(errors, verbose=True)
    assert 6 == ec
    expected = [
        'Command completed with 6 errors or warnings.',
        'CRITICAL: msg1',
        'ERROR: msg2',
        'INFO: msg3',
        'WARNING: msg4',
        'DEBUG: msg4',
        'NOTSET: msg4']
    assert expected == emsgs


class TestFilterError(unittest.TestCase):

    def test_filter_errors_default(self):
        errors = [
            Error(CRITICAL, 'msg1'),
            Error(ERROR, 'msg2'),
            Error(INFO, 'msg3'),
            Error(WARNING, 'msg4'),
            Error(DEBUG, 'msg4'),
            Error(NOTSET, 'msg4'),
        ]
        expected = [
            Error(CRITICAL, 'msg1'),
            Error(ERROR, 'msg2'),
            Error(WARNING, 'msg4'),
        ]
        assert expected == cmd.filter_errors(errors)

    def test_filter_errors_with_min(self):
        errors = [
            Error(CRITICAL, 'msg1'),
            Error(ERROR, 'msg2'),
            Error(INFO, 'msg3'),
            Error(WARNING, 'msg4'),
            Error(DEBUG, 'msg4'),
            Error(NOTSET, 'msg4'),
        ]
        expected = [
            Error(CRITICAL, 'msg1'),
        ]
        assert expected == cmd.filter_errors(errors, CRITICAL)

    def test_filter_errors_no_errors(self):
        errors = [
            Error(INFO, 'msg3'),
            Error(DEBUG, 'msg4'),
            Error(NOTSET, 'msg4'),
        ]
        assert [] == cmd.filter_errors(errors)

    def test_filter_errors_none(self):
        assert [] == cmd.filter_errors([])


class TestParseKeyValues(unittest.TestCase):

    def test_parse_key_values_empty(self):
        assert ({}, []) == cmd.parse_key_values([])
        assert ({}, []) == cmd.parse_key_values(None)

    def test_parse_key_values_simple(self):
        test = [
            'key=value',
            'This=THat',
            'keY=bar',
        ]
        expected = {
            'key': 'bar',
            'this': 'THat'
            }
        keyvals, errors = cmd.parse_key_values(test)
        assert expected == keyvals
        assert not errors

    def test_parse_key_values_with_errors(self):
        test = [
            'key',
            '=THat',
            'keY=',
            'FOO=bar'
        ]
        expected = {
            'foo': 'bar',
        }
        keyvals, errors = cmd.parse_key_values(test)
        assert expected == keyvals
        expected = [
            'missing <key> in "=THat".',
            'missing <value> in "keY=".',
            'missing <value> in "key".'
        ]
        assert expected == errors

###############################################################################
# Run full cli command
###############################################################################


def check_about_stdout(options, expected_loc, regen=False):
    """
    Run the about command with the `options` list of options. Assert that
    command success and that the stdout is equal to the `expected_loc` test file
    content.
    """
    result = run_about_command_test_click(options)
    if regen:
        expected_file = get_test_loc(expected_loc, must_exists=False)
        with open(expected_file, 'w') as ef:
            ef.write(result.output)

    expected_file = get_test_loc(expected_loc, must_exists=True)
    with open(expected_file, 'r') as ef:
        expected = ef.read()

    assert expected.splitlines(False) == result.output.splitlines(False)


def test_about_help_text():
    check_about_stdout(['--help'], 'test_cmd/help/about_help.txt', regen=False)


def test_about_inventory_help_text():
    check_about_stdout(
        ['inventory', '--help'],
        'test_cmd/help/about_inventory_help.txt', regen=False)


def test_about_gen_help_text():
    check_about_stdout(
        ['gen', '--help'],
        'test_cmd/help/about_gen_help.txt', regen=False)


def test_about_gen_license_help_text():
    check_about_stdout(
        ['gen-license', '--help'],
        'test_cmd/help/about_gen_license_help.txt', regen=False)

def test_about_check_help_text():
    check_about_stdout(
        ['check', '--help'],
        'test_cmd/help/about_check_help.txt', regen=False)


def test_about_attrib_help_text():
    check_about_stdout(
        ['attrib', '--help'],
        'test_cmd/help/about_attrib_help.txt', regen=False)


def test_about_command_fails_with_an_unknown_subcommand():
    test_dir = get_temp_dir()
    result = run_about_command_test_click(['foo', test_dir], expected_rc=2)
    assert 'Error: No such command \'foo\'.' in result.output


def test_about_inventory_command_can_run_minimally_without_error():
    test_dir = get_test_loc('test_cmd/repository-mini')
    result = get_temp_file()
    run_about_command_test_click(['inventory', test_dir, result])


def test_about_gen_command_can_run_minimally_without_error():
    test_inv = get_test_loc('test_cmd/geninventory.csv')
    gen_dir = get_temp_dir()
    run_about_command_test_click(['gen', test_inv, gen_dir])


def test_about_attrib_command_can_run_minimally_without_error():
    test_dir = get_test_loc('test_cmd/repository-mini')
    result = get_temp_file()
    run_about_command_test_click(['attrib', test_dir, result])


def test_about_transform_command_can_run_minimally_without_error():
    test_file = get_test_loc('test_cmd/transform.csv')
    result = get_temp_file('file_name.csv')
    run_about_command_test_click(['transform', test_file, result])


def test_about_transform_help_text():
    check_about_stdout(
        ['transform', '--help'],
        'test_cmd/help/about_transform_help.txt', regen=False)


def test_about_transform_expanded_help_text():
    check_about_stdout(
        ['transform', '--help-format'],
        'test_cmd/help/about_transform_config_help.txt', regen=False)



#!/usr/bin/env python
# -*- coding: utf8 -*-

# ============================================================================
#  Copyright (c) nexB Inc. http://www.nexb.com/ - All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ============================================================================

from collections import OrderedDict
import unittest

from testing_utils import get_test_loc

from attributecode.transform import check_duplicate_fields
from attributecode.transform import transform_data
from attributecode.transform import normalize_dict_data
from attributecode.transform import strip_trailing_fields_csv
from attributecode.transform import strip_trailing_fields_json
from attributecode.transform import Transformer
from attributecode.transform import read_csv_rows, read_excel, read_json
from attributecode.transform import transform_csv, transform_excel, transform_json


class TransformTest(unittest.TestCase):

    def test_transform_data_new_col(self):
        data = [OrderedDict([(u'Directory/Filename', u'/tmp/test.c'), (u'Component', u'test.c'),
                             (u'version', '1'), (u'notes', u'test'), (u'temp', u'foo')])]
        configuration = get_test_loc('test_transform/configuration_new_cols')
        transformer = Transformer.from_file(configuration)

        data, err = transform_data(data, transformer)

        expected_data = [dict(OrderedDict([(u'path', u'/tmp/test.c'),
                                           (u'about_resource', u'/tmp/test.c'),
                                           (u'name', u'test.c'), (u'version', u'1'),
                                           (u'notes', u'test'), (u'temp', u'foo')]))]
        assert len(data) == len(expected_data)
        for d in data:
            assert dict(d) in expected_data

    def test_transform_data(self):
        data = [OrderedDict([(u'Directory/Filename', u'/tmp/test.c'),
                             (u'Component', u'test.c'), (u'version', u'1'),
                             (u'notes', u'test'), (u'temp', u'foo')])]
        configuration = get_test_loc('test_transform/configuration')
        transformer = Transformer.from_file(configuration)

        data, err = transform_data(data, transformer)

        expect_name = [u'about_resource', u'name', u'version']
        expected_data = [dict(OrderedDict([(u'about_resource', u'/tmp/test.c'), (u'name', u'test.c'), (u'version', u'1')]))]

        assert len(data) == len(expected_data)
        for d in data:
            assert dict(d) in expected_data

    def test_transform_data_mutli_rows(self):
        data = [OrderedDict([(u'Directory/Filename', u'/tmp/test.c'), (u'Component', u'test.c'), (u'Confirmed Version', u'v0.01')]),
                OrderedDict([(u'Directory/Filename', u'/tmp/tmp.h'), (u'Component', u'tmp.h'), (u'Confirmed Version', None)])]
        configuration = get_test_loc('test_transform/configuration2')
        transformer = Transformer.from_file(configuration)

        data, err = transform_data(data, transformer)

        expect_name = [u'about_resource', u'name', u'version']
        expected_data = [dict(OrderedDict([(u'about_resource', u'/tmp/test.c'), (u'name', u'test.c'), (u'version', u'v0.01')])),
                         dict(OrderedDict([(u'about_resource', u'/tmp/tmp.h'), (u'name', u'tmp.h'), (u'version', None)]))]

        assert len(data) == len(expected_data)
        for d in data:
            assert dict(d) in expected_data

    def test_normalize_dict_data_scancode(self):
        test_file = get_test_loc('test_transform/input_scancode.json')
        json_data = read_json(test_file)
        data = normalize_dict_data(json_data)
        expected_data = [OrderedDict([(u'path', u'samples'),
                                 (u'type', u'directory'),
                                 (u'name', u'samples'),
                                 (u'base_name', u'samples'),
                                 (u'extension', u''), (u'size', 0),
                                 (u'date', None), (u'sha1', None), (u'md5', None),
                                 (u'mime_type', None), (u'file_type', None),
                                 (u'programming_language', None),
                                 (u'is_binary', False), (u'is_text', False),
                                 (u'is_archive', False), (u'is_media', False),
                                 (u'is_source', False), (u'is_script', False),
                                 (u'licenses', []), (u'license_expressions', []),
                                 (u'copyrights', []), (u'holders', []),
                                 (u'authors', []), (u'packages', []),
                                 (u'emails', []), (u'urls', []),
                                 (u'files_count', 33), (u'dirs_count', 10),
                                 (u'size_count', 1161083), (u'scan_errors', [])])]
        assert data == expected_data

    def test_normalize_dict_data_json(self):
        json_data = OrderedDict([(u'Directory/Filename', u'/aboutcode-toolkit/'),
                                 (u'Component', u'AboutCode-toolkit'),
                                 (u'version', u'1.2.3'), (u'note', u'test'),
                                 (u'temp', u'foo')])
        data = normalize_dict_data(json_data)
        expected_data = [OrderedDict([(u'Directory/Filename', u'/aboutcode-toolkit/'),
                                      (u'Component', u'AboutCode-toolkit'),
                                      (u'version', u'1.2.3'), (u'note', u'test'),
                                      (u'temp', u'foo')])]
        assert data == expected_data

    def test_normalize_dict_data_json_array(self):
        json_data = [OrderedDict([(u'Directory/Filename', u'/aboutcode-toolkit/'),
                    (u'Component', u'AboutCode-toolkit'),
                    (u'version', u'1.0'), (u'temp', u'fpp')]),
                    OrderedDict([(u'Directory/Filename', u'/aboutcode-toolkit1/'),
                    (u'Component', u'AboutCode-toolkit1'),
                    (u'version', u'1.1'), (u'temp', u'foo')])]
        data = normalize_dict_data(json_data)
        expected_data = [OrderedDict([(u'Directory/Filename', u'/aboutcode-toolkit/'),
                        (u'Component', u'AboutCode-toolkit'),
                        (u'version', u'1.0'), (u'temp', u'fpp')]),
                        OrderedDict([(u'Directory/Filename', u'/aboutcode-toolkit1/'),
                        (u'Component', u'AboutCode-toolkit1'),
                        (u'version', u'1.1'),
                        (u'temp', u'foo')])]
        assert data == expected_data

    def test_check_duplicate_fields(self):
        field_name = ['path', 'name', 'path', 'version']
        expected = ['path']
        dups = check_duplicate_fields(field_name)
        assert dups == expected

    def test_strip_trailing_fields_csv(self):
        test = [u'about_resource', u'name ', u' version ']
        expected = [u'about_resource', u'name', u'version']
        result = strip_trailing_fields_csv(test)
        assert result == expected

    def test_strip_trailing_fields_json(self):
        test = [OrderedDict([(u'about_resource', u'/this.c'), (u'name ', u'this.c'), (u' version ', u'0.11.0')])]
        expected = [OrderedDict([(u'about_resource', u'/this.c'), (u'name', u'this.c'), (u'version', u'0.11.0')])]
        result = strip_trailing_fields_json(test)
        assert result == expected

    def test_read_excel(self):
        test_file = get_test_loc('test_transform/simple.xlsx')
        error, data = read_excel(test_file)
        assert not error
        expected = [OrderedDict([('about_resource', '/test.c'), ('name', 'test.c'), ('license_expression', 'mit')]),
                    OrderedDict([('about_resource', '/test2.c'), ('name', 'test2.c'), ('license_expression', 'mit and apache-2.0')])]
        assert data == expected

    def test_read_csv_rows(self):
        test_file = get_test_loc('test_transform/simple.csv')
        data = read_csv_rows(test_file)
        expected = [['about_resource', 'name', 'license_expression'],
                    ['/test.c', 'test.c', 'mit'],
                    ['/test2.c', 'test2.c', 'mit and apache-2.0']]
        assert list(data) == expected

    def test_transform_csv(self):
        test_file = get_test_loc('test_transform/input.csv')
        data, err = transform_csv(test_file)
        expected = [{'Directory/Filename': '/aboutcode-toolkit/',
                     'Component': 'AboutCode-toolkit',
                     'Confirmed Version': '123', 'notes': ''}]
        assert len(err) == 0
        assert data == expected

    def test_transform_excel(self):
        test_file = get_test_loc('test_transform/input.xlsx')
        data, err = transform_excel(test_file)
        expected = [OrderedDict([('Directory/Filename', '/aboutcode-toolkit/'),
                                 ('Component', 'AboutCode-toolkit'),
                                 ('Confirmed Version', 123), ('notes', '')])]
        assert len(err) == 0
        assert data == expected

    def test_transform_json(self):
        test_file = get_test_loc('test_transform/input.json')
        data, err = transform_json(test_file)
        expected = [{'Directory/Filename': '/aboutcode-toolkit/',
                     'Component': 'AboutCode-toolkit',
                     'Confirmed Version': '123', 'notes': ''}]
        assert len(err) == 0
        assert data == expected


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) nexB Inc. and others. All rights reserved.
# ScanCode is a trademark of nexB Inc.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/skeleton for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import itertools
import os
import sys

import click

import utils_thirdparty
import utils_requirements

TRACE = False
TRACE_DEEP = False


@click.command()
@click.option(
    "-r",
    "--requirements",
    "requirements_files",
    type=click.Path(exists=True, readable=True, path_type=str, dir_okay=False),
    metavar="REQUIREMENT-FILE",
    multiple=True,
    required=False,
    help="Path to pip requirements file(s) listing thirdparty packages.",
)
@click.option(
    "--spec",
    "--specifier",
    "specifiers",
    type=str,
    metavar="SPECIFIER",
    multiple=True,
    required=False,
    help="Thirdparty package name==version specification(s) as in django==1.2.3. "
    "With --latest-version a plain package name is also acceptable.",
)
@click.option(
    "-l",
    "--latest-version",
    is_flag=True,
    help="Get the latest version of all packages, ignoring any specified versions.",
)
@click.option(
    "-d",
    "--dest",
    "dest_dir",
    type=click.Path(exists=True, readable=True, path_type=str, file_okay=False),
    metavar="DIR",
    default=utils_thirdparty.THIRDPARTY_DIR,
    show_default=True,
    help="Path to the detsination directory where to save downloaded wheels, "
    "sources, ABOUT and LICENSE files..",
)
@click.option(
    "-w",
    "--wheels",
    is_flag=True,
    help="Download wheels.",
)
@click.option(
    "-s",
    "--sdists",
    is_flag=True,
    help="Download source sdists tarballs.",
)
@click.option(
    "-p",
    "--python-version",
    "python_versions",
    type=click.Choice(utils_thirdparty.PYTHON_VERSIONS),
    metavar="PYVER",
    default=utils_thirdparty.PYTHON_VERSIONS,
    show_default=True,
    multiple=True,
    help="Python version(s) to use for wheels.",
)
@click.option(
    "-o",
    "--operating-system",
    "operating_systems",
    type=click.Choice(utils_thirdparty.PLATFORMS_BY_OS),
    metavar="OS",
    default=tuple(utils_thirdparty.PLATFORMS_BY_OS),
    multiple=True,
    show_default=True,
    help="OS(ses) to use for wheels: one of linux, mac or windows.",
)
@click.option(
    "--index-url",
    "index_urls",
    type=str,
    metavar="INDEX",
    default=utils_thirdparty.PYPI_INDEX_URLS,
    show_default=True,
    multiple=True,
    help="PyPI index URL(s) to use for wheels and sources, in order of preferences.",
)
@click.option(
    "--use-cached-index",
    is_flag=True,
    help="Use on disk cached PyPI indexes list of packages and versions and do not refetch if present.",
)
@click.help_option("-h", "--help")
def fetch_thirdparty(
    requirements_files,
    specifiers,
    latest_version,
    dest_dir,
    python_versions,
    operating_systems,
    wheels,
    sdists,
    index_urls,
    use_cached_index,
):
    """
    Download to --dest THIRDPARTY_DIR the PyPI wheels, source distributions,
    and their ABOUT metadata, license and notices files.

    Download the PyPI packages listed in the combination of:
    - the pip requirements --requirements REQUIREMENT-FILE(s),
    - the pip name==version --specifier SPECIFIER(s)
    - any pre-existing wheels or sdsists found in --dest-dir THIRDPARTY_DIR.

    Download wheels with the --wheels option for the ``--python-version``
    PYVER(s) and ``--operating_system`` OS(s) combinations defaulting to all
    supported combinations.

    Download sdists tarballs with the --sdists option.

    Generate or Download .ABOUT, .LICENSE and .NOTICE files for all the wheels
    and sources fetched.

    Download from the provided PyPI simple --index-url INDEX(s) URLs.
    """
    if not (wheels or sdists):
        print("Error: one or both of --wheels  and --sdists is required.")
        sys.exit(1)

    print(f"COLLECTING REQUIRED NAMES & VERSIONS FROM {dest_dir}")

    existing_packages_by_nv = {
        (package.name, package.version): package
        for package in utils_thirdparty.get_local_packages(directory=dest_dir)
    }

    required_name_versions = set(existing_packages_by_nv.keys())

    for req_file in requirements_files:
        nvs = utils_requirements.load_requirements(
            requirements_file=req_file,
            with_unpinned=latest_version,
        )
        required_name_versions.update(nvs)

    for specifier in specifiers:
        nv = utils_requirements.get_required_name_version(
            requirement=specifier,
            with_unpinned=latest_version,
        )
        required_name_versions.add(nv)

    if latest_version:
        names = set(name for name, _version in sorted(required_name_versions))
        required_name_versions = {(n, None) for n in names}

    if not required_name_versions:
        print("Error: no requirements requested.")
        sys.exit(1)

    if TRACE_DEEP:
        print("required_name_versions:")
        for n, v in required_name_versions:
            print(f"    {n} @ {v}")

    # create the environments matrix we need for wheels
    environments = None
    if wheels:
        evts = itertools.product(python_versions, operating_systems)
        environments = [utils_thirdparty.Environment.from_pyver_and_os(pyv, os) for pyv, os in evts]

    # Collect PyPI repos
    repos = []
    for index_url in index_urls:
        index_url = index_url.strip("/")
        existing = utils_thirdparty.DEFAULT_PYPI_REPOS_BY_URL.get(index_url)
        if existing:
            existing.use_cached_index = use_cached_index
            repos.append(existing)
        else:
            repo = utils_thirdparty.PypiSimpleRepository(
                index_url=index_url,
                use_cached_index=use_cached_index,
            )
            repos.append(repo)

    wheels_fetched = []
    wheels_not_found = []

    sdists_fetched = []
    sdists_not_found = []

    for name, version in sorted(required_name_versions):
        nv = name, version
        print(f"Processing: {name} @ {version}")
        if wheels:
            for environment in environments:
                if TRACE:
                    print(f"  ==> Fetching wheel for envt: {environment}")
                fwfns = utils_thirdparty.download_wheel(
                    name=name,
                    version=version,
                    environment=environment,
                    dest_dir=dest_dir,
                    repos=repos,
                )
                if fwfns:
                    wheels_fetched.extend(fwfns)
                else:
                    wheels_not_found.append(f"{name}=={version} for: {environment}")
                    if TRACE:
                        print(f"      NOT FOUND")

        if sdists:
            if TRACE:
                print(f"  ==> Fetching sdist: {name}=={version}")
            fetched = utils_thirdparty.download_sdist(
                name=name,
                version=version,
                dest_dir=dest_dir,
                repos=repos,
            )
            if fetched:
                sdists_fetched.append(fetched)
            else:
                sdists_not_found.append(f"{name}=={version}")
                if TRACE:
                    print(f"      NOT FOUND")

    if wheels and wheels_not_found:
        print(f"==> MISSING WHEELS")
        for wh in wheels_not_found:
            print(f"  {wh}")

    if sdists and sdists_not_found:
        print(f"==> MISSING SDISTS")
        for sd in sdists_not_found:
            print(f"  {sd}")

    print(f"==> FETCHING OR CREATING ABOUT AND LICENSE FILES")
    utils_thirdparty.fetch_abouts_and_licenses(dest_dir=dest_dir, use_cached_index=use_cached_index)
    utils_thirdparty.clean_about_files(dest_dir=dest_dir)

    # check for problems
    print(f"==> CHECK FOR PROBLEMS")
    utils_thirdparty.find_problems(
        dest_dir=dest_dir,
        report_missing_sources=sdists,
        report_missing_wheels=wheels,
    )


if __name__ == "__main__":
    fetch_thirdparty()
